# Copyright (C) 2018 David Kaplan
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from __future__ import print_function
import numpy as np
from astropy.table import Table
from astropy import constants as c, units as u
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS
import astropy
import healpy
# import cPickle as pickle
import pickle

def inside(X,Y,x0,y0):
    """
    result=inside(X,Y,x0,y0)
    see if the points (x0,y0) is inside the convex polygon defined by (X,Y)
    X,Y and x0,y0 should be 1D arrays
    based on:
    http://demonstrations.wolfram.com/AnEfficientTestForAPointToBeInAConvexPolygon/
    """
    X=X-np.array([x0]).T
    # account for wraps in RA
    if (X>180).any():
        X[X>=180]-=360
    if (X<-180).any():
        X[X<=-180]+=360
    # we won't be able to deal with wraps in Dec
    # so we don't want to do this for very close to the poles
    Y=Y-np.array([y0]).T
    n=X.shape[1]
    aplus=np.ones((X.shape[0]),dtype=np.bool)
    aminus=np.ones((X.shape[0]),dtype=np.bool)
    for i1 in range(n):
        i2=(i1+1) % n
        a=X[:,i2]*Y[:,i1]-X[:,i1]*Y[:,i2]
        aplus=aplus & (a>0)
        aminus=aplus & (a<0)
    return aplus | aminus

def order(X,Y,x0,y0):
    """
    result=order(X,Y,x0,y0)

    X,Y and x0,y0 should be 1D arrays
    based on:
    http://demonstrations.wolfram.com/AnEfficientTestForAPointToBeInAConvexPolygon/
    """
    X=X-x0
    # account for wraps in RA
    if (X>180).any():
        X[X>=180]-=360
    if (X<-180).any():
        X[X<=-180]+=360
    # we won't be able to deal with wraps in Dec
    # so we don't want to do this for very close to the poles
    Y=Y-y0
    angles=np.arctan2(Y,X)
    return np.argsort(angles)

class ZTFtile():

    def __init__(self, RA, Dec, config_struct, number=None):
        """
        z=ZTFtile(RA, Dec, number)
        creates a ZTFtile object centered at the given coordinates
        coordinates can be Quantities (with units) or floats (assumed degrees)

        """
        quadrantsFile = config_struct["quadrantsFile"]

        t=Table.read(quadrantsFile,format='ascii.commented_header')
        # Cartesian coordinates in the plane of projection
        self.quadrant_x=t['DX']*u.deg
        self.quadrant_y=t['DY']*u.deg
        self.quadrant_size=np.array([3072,3080])
        self.quadrant_scale=1*u.arcsec

        if isinstance(RA, astropy.units.quantity.Quantity):            
            self.RA=RA
        else:
            self.RA=RA*u.deg
        if isinstance(Dec, astropy.units.quantity.Quantity):            
            self.Dec=Dec
        else:
            self.Dec=Dec*u.deg

        self.number=number

        self.quadrant_RA,self.quadrant_Dec=self.quadrant_centers()
        
        self._wcs=[]
        for i in range(len(self.quadrant_RA)):
            self._wcs.append(WCS(naxis=2))
            self._wcs[-1].wcs.crpix=(self.quadrant_size+1.0)/2
            self._wcs[-1].wcs.ctype=['RA---TAN','DEC--TAN']
            self._wcs[-1].wcs.crval=[self.quadrant_RA[i].value,
                                     max(min(self.quadrant_Dec[i].value,90),-90)]
            self._wcs[-1].wcs.cd=[[-self.quadrant_scale.to(u.deg).value,0],
                                  [0,-self.quadrant_scale.to(u.deg).value]]
    def quadrant_centers(self):
        """
        alpha,delta=quadrant_centers()
        return celestial coordinates for the centers of each quadrant
        given the pointing center RA,Dec
        """
        # convert to Native longitude (phi) and latitude (theta)
        # need intermediate step (Rtheta) to deal with TAN projection
        # Calabretta & Greisen (2002), Eqn. 14,15
        phi=np.arctan2(self.quadrant_x,-self.quadrant_y)
        Rtheta=np.sqrt(self.quadrant_x**2+self.quadrant_y**2)
        # now to theta using the TAN projection
        # Calabrett & Greisen (2002), Eqn. 55
        theta=np.arctan2(1,Rtheta.to(u.rad).value)*u.rad
        
        # central point of projection (native pole)
        alphap=self.RA
        deltap=self.Dec
        # Native longitude/latitude of fiducial point
        # appropriate for zenithal projections including TAN
        phi0=0*u.deg
        theta0=90*u.deg
        # Native longitude/latitue of celestial pole
        # for delta0<theta0 then phip should be 180 deg
        # and theta0=90 deg
        phip=180*u.deg
        thetap=self.Dec
        
        # Celestial longitude/latitude
        # Calabretta & Greisen (2002), Eqn. 2
        alpha=alphap+np.arctan2(-np.cos(theta)*np.sin(phi-phip),
                                np.sin(theta)*np.cos(deltap)-np.cos(theta)*np.sin(deltap)*np.cos(phi-phip))
        
        delta=(np.arcsin(np.sin(theta)*np.sin(deltap)+np.cos(theta)*np.cos(deltap)*np.cos(phi-phip))).to(u.deg)
        alpha[alpha<0*u.deg]+=360*u.deg
        return alpha,delta

    def _inside(self, RA, Dec):
        """
        _inside(RA,Dec)
        returns whether a given RA,Dec (floats in degrees) is inside the FOV
        """
        on=np.zeros_like(RA)
        for i in range(len(self._wcs)):
            footprint=self._wcs[i].calc_footprint(axes=self.quadrant_size)
            on+=inside(footprint[:,0],footprint[:,1],RA,Dec)
            
        return (on>0)

    def _inside_nogaps(self, RA, Dec):
        """
        _inside_nogaps(RA,Dec)
        returns whether a given RA,Dec (floats in degrees) is inside the rough FOV
        ignores chipgaps
        """
        # get all of the corners of the quadrants
        footprint=np.zeros((4*len(self._wcs),2))
        for i in range(len(self._wcs)):
            footprint[(4*i):(4*(i+1)),:]=self._wcs[i].calc_footprint(axes=self.quadrant_size)
        corners=SkyCoord(footprint[:,0]*u.deg,footprint[:,1]*u.deg)
        # distances from these corners to the central pointing position
        corner_distances=corners.separation(SkyCoord(self.RA,self.Dec))
        # find the 4 furthest points - these will define the outer boundaries
        outer_corners=footprint[np.argsort(corner_distances)[-4:],:]
        # sort them by angle to make sure they are convex
        corner_order=order(outer_corners[:,0],outer_corners[:,1],
                           self.RA.value,self.Dec.value)[::-1]
        outer_corners=outer_corners[corner_order,:]
        on=inside(outer_corners[:,0],outer_corners[:,1],RA,Dec)
        return (on>0)

    def inside(self, *args):
        """
        inside(*args)
        returns whether a given coordinate is inside the FOV
        coordinate can be SkyCoord
        or separate RA,Dec
        RA,Dec can be Quantity (with units) or floats (in degrees)
        """

        if len(args)==1:
            if isinstance(args[0],astropy.coordinates.sky_coordinate.SkyCoord):
                return self._inside(args[0].ra.value,args[0].dec.value)
        elif len(args)==2:
            if isinstance(args[0],astropy.units.quantity.Quantity):
                return self._inside(args[0].value,args[1].value)
            else:
                return self._inside(args[0],args[1])
            
    def inside_nogaps(self, *args):
        """
        inside_nogaps(*args)
        returns whether a given coordinate is inside the rough FOV
        ignoring the chip gaps
        coordinate can be SkyCoord
        or separate RA,Dec
        RA,Dec can be Quantity (with units) or floats (in degrees)
        """

        if len(args)==1:
            if isinstance(args[0],astropy.coordinates.sky_coordinate.SkyCoord):
                return self._inside(args[0].ra.value,args[0].dec.value)
        elif len(args)==2:
            if isinstance(args[0],astropy.units.quantity.Quantity):
                return self._inside_nogaps(args[0].value,args[1].value)
            else:
                return self._inside_nogaps(args[0],args[1])

class HP_coverage:

    def __init__(self, hpfile, nside=256, verbose=False):
        """
        HP_coverage(hpfile, nside=256, verbose=False)
        healpix coverage map
        input healpix file and desired nside for degradation (if needed)
        """

        self.nside=nside
        self.npix=healpy.nside2npix(self.nside)

        # read the healpix map
        if verbose:
            print('Loading %s' % hpfile)
        b,h=healpy.read_map(hpfile,h=True,nest=False,verbose=verbose)
        # convert the header to a dictionary
        h={x[0]: x[1] for x in h}
        self.nside_in=h['NSIDE']
        self.npix_in=healpy.nside2npix(self.nside_in)
        if verbose:
            print('Read %s with NSIDE=%d' % (hpfile, self.nside_in))

        # degrade if necessary
        if self.nside < self.nside_in:
            if verbose:
                print('Converting to NSIDE=%d' % self.nside)
            self.hpmap=healpy.ud_grade(b, self.nside, power=-2)
        else:
            self.hpmap=b
            self.nside=self.nside_in
            self.npix=healpy.nside2npix(self.nside)

        # coordinates for each pixel in the map
        RA,Dec=healpy.pix2ang(self.nside, np.arange(self.npix),lonlat=True)
        self.RA=RA*u.deg
        self.Dec=Dec*u.deg

    def tile_coverage(self, ztftiles):
        """
        covered=tile_coverage(ztftiles)
        returns an array the same size as the healpix map
        composed of booleans to show which pixels are covered
        ztftiles is list of ZTFtile objects()
        """

        try:
            l=len(ztftiles)
        except:
            ztftiles=[ztftiles]

        covered=np.zeros(len(self.hpmap),dtype=np.bool)
        for itile in range(len(ztftiles)):
            covered[ztftiles[itile].inside(self.RA,self.Dec)]=True
        return covered
    
    def tile_probability(self, ztftiles):
        """
        probability=tile_probability(tile_RA,tile_Dec)
        ztftiles is list of ZTFtile objects()
        returns total covered probability
        """

        covered=self.tile_coverage(ztftiles)
        return (self.hpmap*covered).sum()

    def get_tile_values(self, hpmap, ztftiles):
        """
        values=get_tile_values(hpmap, ztftiles)
        gets the values used for sorting the tile contributions
        this is implemented as the integral of the probability for the given map over each tile
        ztftiles is list of ZTFtile objects()
        """
        
        values=np.zeros(len(ztftiles))
        for itile in range(len(ztftiles)):
            values[itile]=hpmap[ztftiles[itile].inside(self.RA,self.Dec)].sum()
        return values
    
    def find_tiles(self, ztftiles, probability_target=0.9, verbose=False):
        """
        tiles,probability=find_tiles(ztftiles, probability_target=0.9, verbose=False)

        give ztftiles (list of ZTFtile objects())

        will find the combination to achieve the total probability target
        """

        # starting values and order
        tile_values=self.get_tile_values(self.hpmap, ztftiles)
        tile_order=np.argsort(tile_values)[::-1]

        tiles=[]
        # this is a copy of the map that we can modify
        hpmapc=np.copy(self.hpmap)
        covered=np.zeros(len(self.hpmap),dtype=np.bool)
        summed_probability=0
        while summed_probability < probability_target and len(tiles)<len(ztftiles):
            # add it to the list
            tiles.append(tile_order[0])

            individual_prob=(hpmapc[ztftiles[tile_order[0]].inside(self.RA,self.Dec)].sum())
            if verbose:
                print('Adding tile %d (%f,%f): individual probability = %.3f, total probability = %.3f' % (tile_order[0],
                                                                                                           ztftiles[tile_order[0]].RA.value,
                                                                                                           ztftiles[tile_order[0]].Dec.value,
                                                                                                           individual_prob,
                                                                                                           summed_probability+individual_prob))
            covered[ztftiles[tile_order[0]].inside(self.RA,self.Dec)]=True
            summed_probability+=individual_prob
            hpmapc[ztftiles[tile_order[0]].inside(self.RA,self.Dec)]=0
                
            # redo the priorities to account for the new 
            # probability values
            tile_values=self.get_tile_values(hpmapc, 
                                             ztftiles)
            # priorities for each of those
            tile_order=np.argsort(tile_values)[::-1]

        return tiles,summed_probability

    def percentile(self, level):
        """
        value=percentile(level)
        determines the value of the map above which there is level of total probabiilty
        """
        m=np.sort(self.hpmap)[::-1]
        msum=np.cumsum(m)
        include=msum<level*m.sum()
        # add one more point just to be sure
        if include.sum()<len(m):
            include[include.sum()]=True
        return m[include][-1]

    def percentile_area(self, level):
        """
        area=percentile_area(level)
        determines the area of the map above which there is level of total probabiilty        
        """
        return (self.hpmap>=self.percentile(level)).sum()*healpy.nside2pixarea(self.nside,degrees=True)*u.deg**2




import os, sys, copy
import numpy as np
import healpy as hp

from VOEventLib.VOEvent import Table, Field, What
from VOEventLib.Vutil import utilityTable, stringVOEvent, VOEventExportClass

import astropy.coordinates
from astropy.time import Time, TimeDelta
import astropy.units as u
import ligo.segments as segments
import gwemopt.utils
import gwemopt.rankedTilesGenerator
import gwemopt.hungarian

def get_altaz_tiles(ras, decs, observatory, obstime):

    # Convert to RA, Dec.
    radecs = astropy.coordinates.SkyCoord(
            ra=np.array(ras)*u.degree, dec=np.array(decs)*u.degree, frame='icrs')

    # Alt/az reference frame at observatory, now
    frame = astropy.coordinates.AltAz(obstime=obstime, location=observatory)
            
    # Transform grid to alt/az coordinates at observatory, now
    altaz = radecs.transform_to(frame)    

    return altaz


def find_tile(exposureids_tile,exposureids,probs, idxs = None, exptimecheckkeys = []):
    # exposureids_tile: {expo id}-> list of the tiles available for observation
    # exposureids: list of tile ids for every exposure it is allocated to observe
    if idxs is not None:
        for idx in idxs:
            if len(exposureids_tile["exposureids"])-1 < idx: continue
            idx2 = exposureids_tile["exposureids"][idx]
            if idx2 in exposureids and not idx2 in exptimecheckkeys:
                idx = exposureids.index(idx2)
                exposureids.pop(idx)
                probs.pop(idx)
                return idx2, exposureids, probs

    findTile = True
    while findTile:
        if not exposureids_tile["probs"]:
            idx2 = -1
            findTile = False
            break
        idx = np.argmax(exposureids_tile["probs"])
        idx2 = exposureids_tile["exposureids"][idx]

        if exposureids:
            if idx2 in exposureids and not idx2 in exptimecheckkeys:
                idx = exposureids.index(idx2)
                exposureids.pop(idx)
                probs.pop(idx)
                findTile = False
            else:
                exposureids_tile["exposureids"].pop(idx)
                exposureids_tile["probs"].pop(idx)
        else:
            findTile = False

    return idx2, exposureids, probs


def find_tile_greedy_slew(current_time, current_ra, current_dec, tilesegmentlists, tileprobs, config_struct, tile_struct, 
        keynames, tileAllocatedTime, exptimecheckkeys = [], idle = 0):
    next_obs = -1
    idx2 = -1
    score_selected = -1
    slew_readout_selected = -1
    explength_selected = -1
    for ii in range(len(tilesegmentlists)): # for every tile
        key = keynames[ii]
        # exclude some tiles
        if keynames[ii] in exptimecheckkeys or np.absolute(tileAllocatedTime[ii]) < 1e-5:
            continue
        # calculate slew readout time
        distance = np.sqrt((tile_struct[key]['ra'] - current_ra)**2 +  (tile_struct[key]['dec'] - current_dec)**2)
        slew_readout = np.max([config_struct['readout'], distance / config_struct['slew_rate']])
        slew_readout = np.max([slew_readout - idle, 0])
        slew_readout = slew_readout / 86400
        for jj in range(len(tilesegmentlists[ii])): # for every segment
            seg = tilesegmentlists[ii][jj]
            if current_time + slew_readout < seg[1]: # if ends later than current + slew
                if current_time + slew_readout >= seg[0]: # if starts earlier than current + slew
                    # calculate the score
                    explength = np.min([seg[1] - current_time - slew_readout, tileAllocatedTime[ii]])
                    score = tileprobs[ii] * explength / (1 + slew_readout)
                    if idx2 == -1 or score > score_selected:
                        idx2 = keynames[ii]
                        score_selected = score
                        slew_readout_selected = slew_readout
                        explength_selected = explength
                elif idx2 == -1: # if starts later than current + slew
                    if next_obs == -1 or next_obs > seg[0]:
                        next_obs = seg[0]
                break
    exp_idle_seg = None
    if idx2 != -1:
        exp_idle_seg = segments.segment(current_time + slew_readout, current_time + slew_readout + explength_selected)
    elif next_obs != -1:
        exp_idle_seg = segments.segment(current_time, next_obs)
    return idx2, slew_readout_selected, exp_idle_seg


def get_order(params, tile_struct, tilesegmentlists, exposurelist, observatory, config_struct):
    '''
    tile_struct: dictionary. key -> struct info. 
    tilesegmentlists: list of lists. Segments for each tile in tile_struct 
        that are available for observation.
    exposurelist: list of segments that the telescope is supposed to be working.
        consecutive segments from the start to the end, with each segment size
        being the exposure time.
    Returns a list of tile indices in the order of observation.
    '''
    keys = tile_struct.keys()

    exposureids_tiles = {}
    first_exposure = np.inf*np.ones((len(keys),))
    last_exposure = -np.inf*np.ones((len(keys),))
    tileprobs = np.zeros((len(keys),))
    tilenexps = np.zeros((len(keys),))
    tileexptime = np.zeros((len(keys),))
    tilefilts = {}
    tileavailable = np.zeros((len(keys),))
    tileavailable_tiles = {} 
    keynames = []

    nexps = 0
    for jj, key in enumerate(keys):
        tileprobs[jj] = tile_struct[key]["prob"]
        tilenexps[jj] = tile_struct[key]["nexposures"]
        tilefilts[key] = copy.deepcopy(tile_struct[key]["filt"])
        tileavailable_tiles[jj] = []
        keynames.append(key) 

        nexps = nexps + tile_struct[key]["nexposures"]

    if "dec_constraint" in config_struct:
        dec_constraint = config_struct["dec_constraint"].split(",")
        dec_min = float(dec_constraint[0])
        dec_max = float(dec_constraint[1])

    for ii in range(len(exposurelist)):
        exposureids_tiles[ii] = {}
        exposureids = []
        probs = []
        for jj, key in enumerate(keys):
            tilesegmentlist = tilesegmentlists[jj]
            if tilesegmentlist.intersects_segment(exposurelist[ii]):
                if tile_struct[key]["prob"] == 0: continue
                if "dec_constraint" in config_struct:
                    print(tile_struct[key]["dec"],dec_min,dec_max)
                    if (tile_struct[key]["dec"] < dec_min) or (tile_struct[key]["dec"] > dec_max): continue
                exposureids.append(key)
                probs.append(tile_struct[key]["prob"])

                first_exposure[jj] = np.min([first_exposure[jj],ii])
                last_exposure[jj] = np.max([last_exposure[jj],ii])
                tileavailable_tiles[jj].append(ii)
                tileavailable[jj] = tileavailable[jj] + 1
        # in every exposure, the tiles available for observation
        exposureids_tiles[ii]["exposureids"] = exposureids # list of tile ids
        exposureids_tiles[ii]["probs"] = probs # the corresponding probs

    exposureids = []
    probs = []
    ras, decs = [], []
    for ii, key in enumerate(keys):
        # tile_struct[key]["nexposures"]: the number of exposures assigned to this tile
        for jj in range(tile_struct[key]["nexposures"]): 
            exposureids.append(key) # list of tile ids for every exposure it is allocated to observe
            probs.append(tile_struct[key]["prob"])
            ras.append(tile_struct[key]["ra"])
            decs.append(tile_struct[key]["dec"])

    idxs = -1*np.ones((len(exposureids_tiles.keys()),))
    filts = ['n'] * len(exposureids_tiles.keys())

    if nexps == 0:
        return idxs, filts    

    if params["scheduleType"] == "airmass_weighted":

        # # first step is to sort the array in order of descending probability
        # indsort = np.argsort(-np.array(probs))
        # probs = np.array(probs)[indsort]
        # ras = np.array(ras)[indsort]
        # decs = np.array(decs)[indsort]
        # exposureids = np.array(exposureids)[indsort] 

        matrix = []
        probmatrix = []

        old_exposureids = exposureids.copy()
        for ii in np.arange(len(exposurelist)):

            # first, check whether the entire patch is above the horizon
            t = Time(exposurelist[ii][0], format='mjd')
            altaz = get_altaz_tiles(ras, decs, observatory, t)
            alts = altaz.alt.degree
            horizon = config_struct["horizon"]
            horizon_mask = alts <= horizon
            airmass = 1 / np.cos((90. - alts) * np.pi / 180.)
            below_horizon_mask = horizon_mask * 10.**100
            airmass = airmass + below_horizon_mask

            # if part of patch below horizon, apply greedy method
            # temporary solution until runtime issue fixed
            if horizon_mask.any() or len(exposureids)==1:
                exptimecheck = np.where(exposurelist[ii][0]-tileexptime <
                    params["mindiff"]/86400.0)[0]
                exptimecheckkeys = [keynames[x] for x in exptimecheck]
                # find_tile finds the tile that covers the largest probablity
                # restricted by availability of tile and timeallocation
                idx2, exposureids, probs = find_tile(exposureids_tiles[ii],exposureids,probs,exptimecheckkeys=exptimecheckkeys)
                
                if idx2 in old_exposureids:
                    idx = old_exposureids.index(idx2)
                    ras.pop(idx)
                    decs.pop(idx)
                    old_exposureids.pop(idx)

                if idx2 in keynames:
                    idx = keynames.index(idx2)
                    tilenexps[idx] = tilenexps[idx] - 1
                    tileexptime[idx] = exposurelist[ii][0]
                    if len(tilefilts[idx2]) > 0:
                        filt = tilefilts[idx2].pop(0)
                        filts[ii] = filt

                idxs[ii] = idx2

            # once the patch has risen, create a matrix of the remaining exposure times and tiles
            elif not horizon_mask.any() and len(exposureids) > 1:
                matrix.append(probs/airmass)
                probmatrix.append(probs * (True^horizon_mask))

            tilematrix = np.array(matrix)

    if params["scheduleType"] == "greedy":
        for ii in np.arange(len(exposurelist)): 
            exptimecheck = np.where(exposurelist[ii][0]-tileexptime <
                                    params["mindiff"]/86400.0)[0]
            exptimecheckkeys = [keynames[x] for x in exptimecheck]

            # find_tile finds the tile that covers the largest probablity
            # restricted by availability of tile and timeallocation
            idx2, exposureids, probs = find_tile(exposureids_tiles[ii],exposureids,probs,exptimecheckkeys=exptimecheckkeys)
            if idx2 in keynames:
                idx = keynames.index(idx2)
                tilenexps[idx] = tilenexps[idx] - 1
                tileexptime[idx] = exposurelist[ii][0]
                if len(tilefilts[idx2]) > 0:
                    filt = tilefilts[idx2].pop(0)
                    filts[ii] = filt
            idxs[ii] = idx2

            if not exposureids: break

    elif params["scheduleType"] == "sear":
        #for ii in np.arange(len(exposurelist)):
        iis = np.arange(len(exposurelist)).tolist()
        while len(iis) > 0: 
            ii = iis[0]
            mask = np.where((ii == last_exposure) & (tilenexps > 0))[0]

            exptimecheck = np.where(exposurelist[ii][0]-tileexptime <
                                    params["mindiff"]/86400.0)[0]
            exptimecheckkeys = [keynames[x] for x in exptimecheck]

            if len(mask) > 0:
                idxsort = mask[np.argsort(tileprobs[mask])]
                idx2, exposureids, probs = find_tile(exposureids_tiles[ii],exposureids,probs,idxs=idxsort,exptimecheckkeys=exptimecheckkeys) 
                last_exposure[mask] = last_exposure[mask] + 1
            else:
                idx2, exposureids, probs = find_tile(exposureids_tiles[ii],exposureids,probs,exptimecheckkeys=exptimecheckkeys)
            if idx2 in keynames:
                idx = keynames.index(idx2)
                tilenexps[idx] = tilenexps[idx] - 1
                tileexptime[idx] = exposurelist[ii][0]
                if len(tilefilts[idx2]) > 0:
                    filt = tilefilts[idx2].pop(0)
                    filts[ii] = filt
            idxs[ii] = idx2 
            iis.pop(0)

            if not exposureids: break

    elif params["scheduleType"] == "weighted":
        for ii in np.arange(len(exposurelist)):
            jj = exposureids_tiles[ii]["exposureids"]
            weights = tileprobs[jj] * tilenexps[jj] / tileavailable[jj]
            weights[~np.isfinite(weights)] = 0.0

            exptimecheck = np.where(exposurelist[ii][0]-tileexptime <
                                    params["mindiff"]/86400.0)[0]
            weights[exptimecheck] = 0.0

            if np.any(weights >= 0):
                idxmax = np.argmax(weights)
                idx2 = jj[idxmax]
                if idx2 in keynames:
                    idx = keynames.index(idx2)
                    tilenexps[idx] = tilenexps[idx] - 1
                    tileexptime[idx] = exposurelist[ii][0]
                    if len(tilefilts[idx2]) > 0:
                        filt = tilefilts[idx2].pop(0)
                        filts[ii] = filt
                idxs[ii] = idx2
            tileavailable[jj] = tileavailable[jj] - 1  

    elif params["scheduleType"] == "airmass_weighted":
        
        tilematrix_mask = tilematrix > 10**(-10)

        if tilematrix_mask.any():
            hungarian = gwemopt.hungarian.Hungarian(tilematrix, profit_matrix=True)
            hungarian.calculate()
            optimal_points = np.array(hungarian.assigned_points)
            # sort the optimal points first on probability
            order = np.argsort(optimal_points[:, 0])
            optimal_points = optimal_points[order]
            max_no_observ = min(tilematrix.shape)

            for jj in range(max_no_observ):
                # idx0 indexes over the time windows, idx1 indexes over the probabilities
                # idx2 gets the exposure id of the tile, used to assign tileexptime and tilenexps
                try:
                    idx0 = optimal_points[jj][1]
                    idx1 = optimal_points[jj][0]
                    idx2 = exposureids[idx1.astype(int)]
                    tileprob = probmatrix[idx0][idx1]
                    if len(tilefilts[idx2]) > 0:
                        filt = tilefilts[idx2].pop(0)
                        filts[jj] = filt
                except: continue

                greedy_tile_idx = len(exposurelist) - max(tilematrix.shape)
                idxs[idx0 + greedy_tile_idx] = idx2


        else: print("The localization is not visible from the site.")

    else:
        print("Scheduling options are greedy/sear/weighted/airmass_weighted, or with _slew.")
        exit(0)

    return idxs, filts


def get_order_slew(params, tile_struct, tilesegmentlists, config_struct):   
    keys = tile_struct.keys() 
    keynames = []
    namekeys = {}
    tilefilts = {}
    tileAllocatedTime = {}
    tileprobs = np.zeros((len(keys),))
    for jj, key in enumerate(keys):
        tileprobs[jj] = tile_struct[key]["prob"]
        tilefilts[key] = copy.deepcopy(tile_struct[key]["filt"])
        keynames.append(key) 
        namekeys[key] = jj
        tileAllocatedTime[key] = tile_struct[key]["exposureTime"] / 86400
    lastObs = np.zeros((len(keys),))

    idxs = []
    filts = []
    exposurelist = []

    if params["scheduleType"] == "greedy_slew":
        current_time = Time(params['gpstime'], format='gps', scale='utc').mjd + np.min(params["Tobs"][::2])
        current_ra = config_struct["latitude"]
        current_dec = config_struct["longitude"]
        idle = 0
        while True:
            # check the time gap since last observation
            exptimecheck = np.where(current_time - lastObs < params["mindiff"]/86400.0)[0]
            exptimecheckkeys = [keynames[x] for x in exptimecheck]
            idx2, slew_readout, exp_idle_seg = find_tile_greedy_slew(current_time, current_ra, current_dec, tilesegmentlists, tileprobs, 
                    config_struct, tile_struct, keynames, tileAllocatedTime, exptimecheckkeys=exptimecheckkeys, idle=idle)
            if idx2 != -1:
                exp_idle_len = exp_idle_seg[1] - exp_idle_seg[0]
                idle = 0
                idxs.append(idx2)
                exposurelist.append(exp_idle_seg)
                filts.append(tilefilts[idx2].pop(0))
                tileAllocatedTime[idx2] = tileAllocatedTime[idx2] - exp_idle_len
                current_time = exp_idle_seg[1]
                current_ra = tile_struct[idx2]["ra"]
                current_dec = tile_struct[idx2]["dec"]
                idx = namekeys[idx2]
                lastObs[idx] = current_time
            elif exp_idle_seg is not None:
                exp_idle_len = exp_idle_seg[1] - exp_idle_seg[0]
                idxs.append(idx2)
                exposurelist.append(exp_idle_seg)
                filts.append('n')
                idle = exp_idle_len
                current_time += exp_idle_len
            else:
                break
    elif params["scheduleType"] == "sear_slew":
        print('sear_slew is not ready yet.')
    elif params["scheduleType"] == "weighted_slew":
        print('weighted_slew is not ready yet.')
    else:
        print("Scheduling options are greedy/sear/weighted, or with _slew.")
        exit(0)

    return idxs, exposurelist, filts


def scheduler(params, config_struct, tile_struct):
    '''
    config_struct: the telescope configurations
    tile_struct: the tiles, contains time allocation information
    '''
    import gwemopt.segments
    #import gwemopt.segments_astroplan
    coverage_struct = {}
    coverage_struct["data"] = np.empty((0,8))
    coverage_struct["filters"] = []
    coverage_struct["ipix"] = []
    coverage_struct["patch"] = []
    coverage_struct["area"] = []

    observatory = astropy.coordinates.EarthLocation(
        lat=config_struct["latitude"]*u.deg, lon=config_struct["longitude"]*u.deg, height=config_struct["elevation"]*u.m)

    segmentlist = config_struct["segmentlist"]
    exposurelist = config_struct["exposurelist"]
    #tilesegmentlists = gwemopt.segments_astroplan.get_segments_tiles(config_struct, tile_struct, observatory, segmentlist)
    tilesegmentlists = []
    keys = tile_struct.keys()
    for key in keys:
        # segments.py: tile_struct[key]["segmentlist"] is a list of segments when the tile is available for observation
        tilesegmentlists.append(tile_struct[key]["segmentlist"]) 
    print("Generating schedule order...")
    if params["scheduleType"].endswith('_slew'):
        keys, exposurelist, filts = get_order_slew(params, tile_struct, tilesegmentlists, config_struct)
    else:
        keys, filts = get_order(params,tile_struct,tilesegmentlists,exposurelist,observatory, config_struct)
    if params["doPlots"]:
        gwemopt.plotting.scheduler(params,exposurelist,keys)

    exposureused = np.where(np.array(keys)>=0)[0]
    coverage_struct["exposureused"] = exposureused
    while len(exposurelist) > 0:
        key, filt = keys[0], filts[0]
        if key == -1:
            keys = keys[1:]
            filts = filts[1:]
            exposurelist = exposurelist[1:]
        else:
            tile_struct_hold = tile_struct[key]
            mjd_exposure_start = exposurelist[0][0]
            nkeys = len(keys)
            for jj in range(nkeys):
                if (keys[jj] == key) and (filts[jj] == filt) and not (nkeys == jj+1):
                    mjd_exposure_end = exposurelist[jj][1]
                elif (keys[jj] == key) and (filts[jj] == filt) and (nkeys == jj+1):
                    mjd_exposure_end = exposurelist[jj][1]
                    nexp = jj + 1
                    keys = []
                    filts = []
                    exposurelist = []
                else:
                    nexp = jj + 1
                    keys = keys[jj:]
                    filts = filts[jj:]
                    exposurelist = exposurelist[jj:]
                    break    
 
            # the (middle) tile observation time is mjd_exposure_mid
            mjd_exposure_mid = (mjd_exposure_start+mjd_exposure_end)/2.0

            # calculate airmass for each tile at the middle of its exposure:
            t = Time(mjd_exposure_mid, format='mjd')
            altaz = get_altaz_tiles(tile_struct_hold["ra"], tile_struct_hold["dec"], observatory, t)
            alt = altaz.alt.degree
            airmass = 1 / np.cos((90. - alt) * np.pi / 180)

            nmag = np.log(nexp) / np.log(2.5)
            mag = config_struct["magnitude"] + nmag
            exposureTime = (mjd_exposure_end-mjd_exposure_start)*86400.0

            coverage_struct["data"] = np.append(coverage_struct["data"],np.array([[tile_struct_hold["ra"],tile_struct_hold["dec"],mjd_exposure_mid,mag,exposureTime,int(key),tile_struct_hold["prob"],airmass]]),axis=0)

            coverage_struct["filters"].append(filt)
            coverage_struct["patch"].append(tile_struct_hold["patch"])
            coverage_struct["ipix"].append(tile_struct_hold["ipix"])
            coverage_struct["area"].append(tile_struct_hold["area"])

    coverage_struct["area"] = np.array(coverage_struct["area"])
    coverage_struct["filters"] = np.array(coverage_struct["filters"])
    coverage_struct["FOV"] = config_struct["FOV"]*np.ones((len(coverage_struct["filters"]),))
    return coverage_struct


def computeSlewReadoutTime(config_struct, coverage_struct):
    slew_rate = config_struct['slew_rate']
    readout = config_struct['readout']
    prev_ra = config_struct["latitude"]
    prev_dec = config_struct["longitude"]
    acc_time = 0
    for dat in coverage_struct['data']:
        slew_readout_time = np.maximum(np.sqrt((prev_ra - dat[0])**2 + (prev_dec - dat[1])**2) / slew_rate, readout)
        acc_time += slew_readout_time
        prev_dec = dat[0]
        prev_ra = dat[1]
    return acc_time

def write_xml(xmlfile,map_struct,coverage_struct,config_struct):

    what = What()

    table = Table(name="data", Description=["The datas of GWAlert"])
    table.add_Field(Field(name=r"grid_id", ucd="", unit="", dataType="int", \
                    Description=["ID of the grid of fov"]))
    table.add_Field(Field(name="field_id", ucd="", unit="", dataType="int",\
                    Description=["ID of the filed"]))
    table.add_Field(
        Field(
            name=r"ra", ucd=r"pos.eq.ra ", unit="deg", dataType="float",
            Description=["The right ascension at center of fov in equatorial coordinates"]
            )
        )
    table.add_Field(
        Field(
            name="dec", ucd="pos.eq.dec ", unit="deg", dataType="float",
            Description=["The declination at center of fov in equatorial coordinates"]
            )
        )
    table.add_Field(
        Field(
            name="ra_width", ucd=" ", unit="deg", dataType="float",
            Description=["Width in RA of the fov"]
            )
        )
    table.add_Field(
        Field(
            name="dec_width", ucd="", unit="deg", dataType="float",
            Description=["Width in Dec of the fov"]
            )
        )
    table.add_Field(
        Field(
            name="prob_sum", ucd="", unit="None", dataType="float",
            Description=["The sum of all pixels in the fov"]
            )
        )
    table.add_Field(
        Field(
            name="observ_time", ucd="", unit="sec", dataType="float",
            Description=["Tile mid. observation time in MJD"]
            )
        )
    table.add_Field(
        Field(
            name="airmass", ucd="", unit="None", dataType="float",
            Description=["Airmass of tile at mid. observation time"]
            )
        )
    table.add_Field(Field(name="priority", ucd="", unit="", dataType="int", Description=[""]))
    table_field = utilityTable(table)
    table_field.blankTable(len(coverage_struct))

    for ii in range(len(coverage_struct["ipix"])):
        data = coverage_struct["data"][ii,:]
        filt = coverage_struct["filters"][ii]
        ipix = coverage_struct["ipix"][ii]
        patch = coverage_struct["patch"][ii]
        FOV = coverage_struct["FOV"][ii]
        area = coverage_struct["area"][ii]

        prob = np.sum(map_struct["prob"][ipix])

        ra, dec = data[0], data[1]
        observ_time, exposure_time, field_id, prob, airmass = data[2], data[4], data[5], data[6], data[7]

        table_field.setValue("grid_id", ii, 0)
        table_field.setValue("field_id", ii, field_id)
        table_field.setValue("ra", ii, ra)
        table_field.setValue("dec", ii, dec)
        table_field.setValue("ra_width", ii, config_struct["FOV"])
        table_field.setValue("dec_width", ii, config_struct["FOV"])
        table_field.setValue("observ_time", ii, observ_time)
        table_field.setValue("airmass", ii, airmass)
        table_field.setValue("prob_sum", ii, prob)
        table_field.setValue("priority", ii, ii)
    table = table_field.getTable()
    what.add_Table(table)
    xml = stringVOEvent(what)
    lines = xml.splitlines()
    linesrep = []
    for line in lines:
        linenew = line.replace(">b'",">").replace("'</","</").replace("=b'","=").replace("'>",">")
        linesrep.append(linenew)
    xmlnew = "\n".join(linesrep)
    fid = open(xmlfile, "w")
    fid.write(xmlnew)
    fid.close()

def summary(params, map_struct, coverage_struct):

    idx50 = len(map_struct["cumprob"])-np.argmin(np.abs(map_struct["cumprob"]-0.50))
    idx90 = len(map_struct["cumprob"])-np.argmin(np.abs(map_struct["cumprob"]-0.90))

    mapfile = os.path.join(params["outputDir"],'map.dat')
    fid = open(mapfile,'w')
    fid.write('%.5f %.5f\n'%(map_struct["pixarea_deg2"]*idx50,map_struct["pixarea_deg2"]*idx90))
    fid.close()
    filts = list(set(coverage_struct["filters"]))
    for telescope in params["telescopes"]:

        schedulefile = os.path.join(params["outputDir"],'schedule_%s.dat'%telescope)
        schedulexmlfile = os.path.join(params["outputDir"],'schedule_%s.xml'%telescope)
        config_struct = params["config"][telescope]

        write_xml(schedulexmlfile,map_struct,coverage_struct,config_struct)

        if (params["tilesType"] == "hierarchical") or (params["tilesType"] == "greedy"):
            fields = np.zeros((params["Ntiles"],len(filts)+2))
        else:
            fields = np.zeros((len(config_struct["tesselation"]),len(filts)+2))

        fid = open(schedulefile,'w')
        for ii in range(len(coverage_struct["ipix"])):
            data = coverage_struct["data"][ii,:]
            filt = coverage_struct["filters"][ii]
            ipix = coverage_struct["ipix"][ii]
            patch = coverage_struct["patch"][ii]
            FOV = coverage_struct["FOV"][ii]
            area = coverage_struct["area"][ii]

            prob = np.sum(map_struct["prob"][ipix])

            ra, dec = data[0], data[1]
            observ_time, exposure_time, field_id, prob, airmass = data[2], data[4], data[5], data[6], data[7]
            fid.write('%d %.5f %.5f %.5f %d %.5f %.5f %s\n'%(field_id,ra,dec,observ_time,exposure_time,prob,airmass,filt))

            idx1 = np.argmin(np.sqrt((config_struct["tesselation"][:,1]-data[0])**2 + (config_struct["tesselation"][:,2]-data[1])**2))
            idx2 = filts.index(filt)
            fields[idx1,0] = config_struct["tesselation"][idx1,0]
            fields[idx1,1] = prob
            fields[idx1,idx2+2] = fields[idx1,idx2+2]+1
        fid.close()

        idx = np.where(fields[:,1]>0)[0]
        fields = fields[idx,:]
        idx = np.argsort(fields[:,1])[::-1]
        fields = fields[idx,:]

        fields_sum = np.sum(fields[:,2:],axis=1)
        idx = np.where(fields_sum >= 2)[0]
        print('%d/%d fields were observed at least twice\n'%(len(idx),len(fields_sum)))

        print('Integrated probability, All: %.5f, 2+: %.5f'%(np.sum(fields[:,1]),np.sum(fields[idx,1])))

        slew_readout_time = computeSlewReadoutTime(config_struct, coverage_struct)
        print('Expected time spent on slewing and readout ' + str(slew_readout_time) + ' s.')

        coveragefile = os.path.join(params["outputDir"],'coverage_%s.dat'%telescope)
        fid = open(coveragefile,'w')
        for field in fields:
            fid.write('%d %.10f '%(field[0],field[1]))
            for ii in range(len(filts)):
                 fid.write('%d '%(field[2+ii]))
            fid.write('\n')
        fid.close()

    summaryfile = os.path.join(params["outputDir"],'summary.dat')
    fid = open(summaryfile,'w')

    gpstime = params["gpstime"]
    event_mjd = Time(gpstime, format='gps', scale='utc').mjd

    tts = np.array([1,7,60])
    for tt in tts:

        radecs = np.empty((0,2))
        mjds_floor = []
        mjds = []
        ipixs = np.empty((0,2))
        cum_prob = 0.0
        cum_area = 0.0

        for ii in range(len(coverage_struct["ipix"])):
            data = coverage_struct["data"][ii,:]
            filt = coverage_struct["filters"][ii]
            ipix = coverage_struct["ipix"][ii]
            patch = coverage_struct["patch"][ii]
            FOV = coverage_struct["FOV"][ii]
            area = coverage_struct["area"][ii]

            prob = np.sum(map_struct["prob"][ipix])

            if data[2] > event_mjd+tt:
                continue

            ipixs = np.append(ipixs,ipix)
            ipixs = np.unique(ipixs).astype(int)

            cum_prob = np.sum(map_struct["prob"][ipixs])
            cum_area = len(ipixs) * map_struct["pixarea_deg2"]
            mjds.append(data[2])
            mjds_floor.append(int(np.floor(data[2])))

        if len(mjds_floor) == 0:
            print("No images after %.1f days..."%tt)
            fid.write('%.1f,-1,-1,-1,-1\n'%(tt))
        else:

            mjds = np.unique(mjds)
            mjds_floor = np.unique(mjds_floor)

            print("After %.1f days..."%tt)
            print("Number of hours after first image: %.5f"%(24*(np.min(mjds)-event_mjd)))
            print("MJDs covered: %s"%(" ".join(str(x) for x in mjds_floor)))
            print("Cumultative probability: %.5f"%cum_prob)
            print("Cumultative area: %.5f degrees"%cum_area)

            fid.write('%.1f,%.5f,%.5f,%.5f,%s\n'%(tt,24*(np.min(mjds)-event_mjd),cum_prob,cum_area," ".join(str(x) for x in mjds_floor)))

    fid.close()





# coding: utf-8

# In[81]:


import numpy as np
import numbers
import sys
import scipy.stats
import time

import matplotlib.pyplot as plot
import scipy.spatial as spatial
from scipy import integrate, interpolate, optimize
import pickle
np.core.arrayprint._line_width = 200  

# In[117]:


def Main(T_tot, readout_time, slew_time, PGW, tau, prob, method):
    #Main(T_tot,readout_time,slew_time,PGW,tau,prob)
    #Main is the main body of the optimisation code
    
    #T_tot: total available telescope time, in hrs
    #readout_time : the time needed to perform readout, scalar.
    
    #slew time: the time needed to slew telescope , scalar.
    #           If the time needed to perform both slew and readout is just the larger
    #           of the two, just set one of these parameters equal to the larger one
    #           and the other equal to zero.
    #
    #area_index: A matrix representing the fields and containing the GW
    #            probabilities. an output of function 'fields'. 
     
    #tau: A vector containing the observation tau from 0.1 seconds to 1e5
    #    seconds. An output of Pem.
    
    #prob: A vector containing the values of Pem at each of the values of
    #     tau. An output of Pem.       
    
    #T: The total Observation Time available, in second.
    
    if isinstance(T_tot, numbers.Number) == False:
        sys.exit('Please enter the total observation times, in hours')
        
    nof = len(PGW)
    message1=['The number of fields is ', str(nof), '\n']
    message1 = ''.join(message1)
    print(message1)
    
    numberoffields = np.linspace(1, nof, nof)
    
    tprob = np.zeros(np.count_nonzero(PGW))
    prob_distri = [[] for i in range(np.count_nonzero(PGW))]
    time_allocation = [[] for i in range(np.count_nonzero(PGW))]
    
    print('Start generating the optimal strategy...\n')
    
    
    
    for nos in range(nof):
        T = T_tot-(readout_time+slew_time)*(nos+1);   #This computes the total time available to observation only.
        text = ['The number of fields being considered now is ', str(nos + 1), '.']
        #print(''.join(text))
        if np.count_nonzero(PGW) >= nos + 1:
            time_allocated, individual_prob, probability = Solver(PGW[0 : nos + 1], nos + 1, tau, prob, T, method)
        else:
            text = text1=['Because the sum of the posterior probabilities within the fields from the ',num2str(nos), ' th field onwards are zero, this analysis aborts.'];
            #print(''.join(text))
            break
        tprob[nos] = probability
        prob_distri[nos] = individual_prob
        time_allocation[nos] = time_allocated
        
        text = ['Provided observing ', str(nos), ' fields, the detection probability of the target EM counterpart is ', str(tprob[nos]), '\n'];
        #print(''.join(text)) 
        
        text='The time assigned to each of these fields is as follows:';
        #print(''.join(text)) 
        #print(time_allocation[nos])
        #print('\n')
    
    maxprob = np.amax(tprob)
    maxindex = np.argwhere(tprob == maxprob).squeeze()
    #print('Therefore, based on the analyses above \n')
    text = ['The highest possibility could be obtained by observing ', str(maxindex + 1), ' fields, and the detection probability is ', str(maxprob), '. \n'];
    #print(''.join(text))
    text = ['The time should be allocated as follows: ', str(time_allocation[maxindex][0:-1])]
    #print(''.join(text))
    
    text = ['The exit condition for optimize.root is:', ]
    
    
    plot.plot(numberoffields, tprob)
    plot.xlabel('The total observed fields k')
    plot.ylabel('$P(D_{EM}|k)$')
    plot.show()
    
    return maxprob, time_allocation[maxindex]


# In[96]:


def Pem(lim_mag, lim_time, N_ref = 9.7847e9, L_min = 4.9370e31, L_max = 4.9370e33, model = '', sample_length = None, pMdM = None, tau = None, Loftau = 61, D_mu = 200.0e6, D_sig = 60.0e6, R = None, p_R = None):
    #tau,prob = lim_mag, lim_time, N_ref = 9.7847e9, L_min = 4.9370e31, L_max = 4.9370e33, model = '', sample_length = None, pMdM = None, tau = None, Loftau = 61, D_mu = 200.0e6, D_sig = 60.0e6, R = None, p_R = None)
    #Pem calculates the values of the P_EM integral 
    
    #lim_mag: the limiting magnitude of the telescope. For kilonovae, the
    #         magnitude should be in R-band.

    #lim_time: the limiting time at which the lim_mag is achieved, in
    #          seconds. For example, if a telescope can achieve apparent an R
    #          band magnitude 21 in 3 minutes, then lim_time is 180 seconds,
    #          and lim_mag is 21.

    #N_ref: the number of photons expected at apparent magnitude equal to
    #       zero. The default value is 9.7847*10^9.  

    #L_min: the minimum peak Luminosity of kilonovae, in w.
    #       The default value is 4.970e31w (M = -8) 

    #L_max: the maximum peak Luminosity of kilonovae, in w. 
    #       The default value is 4.970e33w (M = -13) from (Barnes &
    #       Kasen, 2013).
    
    # model: the model that determines whether telescopes will see a target given
    #        a specific observation time. 
    #        'Poisson': Poisson distribution. The function will use a Poisson distribution
    #             to determines how many photons will be seen at a given observation time
    #             lambda is the expected number of photons per second per unit area given 
    #             a peak luminosity
    #        '': The function will use a scaling equation to compute the observation time 
    #           needed to achieve a detection given a magnitude. 
    
    # sample_length: the number of points in distance and magnitudes(luminosity) spaces for 
    #               calculating the integral over magnitude and distance (Eq. 5b in 
    #               DOI:10.3847/1538-4357/834/1/84.
    #               The larger the value, the more accurate the intergral is.
    #               the default value is 100. 
    
    # pMdM: p(M) dM where p(M) is the probability of M. 
    #        A vector the same length as sample_length.
    #        The default is that derived from the prior on peak luminosity ( ~1/sqrt(L) )
    #        If this is a user input value, L_min and L_max will not be used.
    
    # tau: The points in observation time space. 
    
    # Loftau: the number of points is observation time space. The default is 61.
    #         tau will then be defined as tau = 10 ^ (-1 + i * 0.1) with i ranging 
    #         from 0 to Loftau - 1.
    
    # D_mu: This function assumes a gaussian distribution to approximate the distribution on 
    #       distance. D_mu is the mean of the distribution, in par sec.
    
    # D_sig: the variance of the gaussian distribution on distance, in par sec.
    
    # R: A vector in distance space. The interval between the elements should be equal
    #    so that dR = R[1] - R[0]. The default is derived from D_min, D_max equal to 1e7
    #    1e9 par sec.
    
    # p_R: the probability distribution density of R. default is gaussian distribution.
    
    start_time = time.time()
        
    if sample_length == None:
        sample_length = 100
        
    L = np.linspace(L_min, L_max, sample_length)
    logL = np.log10(L)
    dL = L[1] - L[0]
    L_sun = 3.85e26
    logL_sun = np.log10(L_sun)
    M = 4.77 - 2.5 * logL + 2.5 * logL_sun
    
    if pMdM == None:
        k = 1.0 / (2.0 * (np.sqrt(L_max) - np.sqrt(L_min)))
        L_prior = k / np.sqrt(L)
        pMdM = L_prior * dL
    elif len(pMdM) != sample_length:
        sys.exit('pMdm has to have length equal to sample_length.') 
    
    
    if p_R is None:
        D_min = 1.0e7
        D_max = 1.0e9
        Rstepsize = (D_max - D_min) / sample_length
        Rsteps = (D_max - D_min) / Rstepsize
        R = np.linspace(D_min, D_max, int(Rsteps))
        
        p_R = scipy.stats.norm(D_mu, D_sig).pdf(R)   
    else:
        Rstepsize = R[1] - R[0]
        Rsteps = len(R)
    
    if tau is None:    
        tau = np.zeros(Loftau)
        for i in range(Loftau):
            tau[i] = 10 ** (-1 + (i) * 0.1)
    else:
        Loftau = len(tau)
    prob = np.zeros(Loftau)
    N_exp = np.zeros((len(R), sample_length))
    
    
    scaling_num = 10 ** (np.log10(N_ref) - lim_mag / 2.5);
    scaling_flux = lim_time * scaling_num 

    message1='Computing the values of P_EM at tau equal to ';
    message2='This calculation will be repeated ';
    message3=' times.';
    message4='There are still ';
    message5=' times to go. \n';
    
    if model == 'Poisson':
        for i in range(Loftau):
            message = [message1, str(tau[i]), 's.']
            print(''.join(message))
            message = [message2, str(Loftau), message3]
            print(''.join(message))
            message = [message4, str(Loftau - i - 1), message5]
            print(''.join(message))

            for dist in range(len(R)):
                for mag in range(sample_length):
                    N_exp[dist][mag] = 10 ** (np.log10(N_ref) - (M[mag] + 5.0 * np.log10(R[dist]) - 5)/2.5)
                    int_value = (1 - scipy.stats.poisson.cdf(scaling_flux, N_exp[dist][mag] * tau[i])) * p_R[dist] * pMdM[mag] * Rstepsize
                    prob[i] += int_value
    else:
        for i in range(Loftau):
            message = [message1, str(tau[i]), 's.']
            print(''.join(message))
            message = [message2, str(Loftau), message3]
            print(''.join(message))
            message = [message4, str(Loftau - i - 1), message5]
            print(''.join(message))

            for dist in range(len(R)):
                for mag in range(sample_length):
                    N_exp[dist][mag] = 10 ** (np.log10(N_ref) - (M[mag] + 5.0 * np.log10(R[dist]) - 5)/2.5)
                    if N_exp[dist][mag] * tau[i] >= scaling_flux:
                        int_value = p_R[dist] * pMdM[mag] * Rstepsize
                        prob[i] += int_value            

    finish_time = time.time() - start_time
    print(''.join(['The calculation has taken ', str(round(finish_time,2)), ' seconds.']))
 
    return tau, prob

def rotx(angle):
    rotR = np.array([[1, 0, 0], [0, np.cos(angle), np.sin(angle)], [0, -np.sin(angle), np.cos(angle)]])
    return rotR

def roty(angle):
    rotR = np.array([[np.cos(angle), 0, -np.sin(angle)], [0, 1, 0], [np.sin(angle), 0, np.cos(angle)]])
    return rotR

def rotz(angle):
    rotR = np.array([[np.cos(angle), np.sin(angle), 0], [-np.sin(angle), np.cos(angle), 0], [0, 0, 1]])
    return rotR


# In[110]:


def Greedy(fov, n, skymap, resolution, level):
    #Greedy(fov,n,skymap,resolution,level)
    #Greedy finds the locations of the fields using a greedy algorithm;

    #fov       : the size of field of view, in deg^2. In our model, a square
    #            fov is assumed such that the length and the width of the fov
    #            is the square root of the fov.

    #n         : the total number of fields that will be chosen by the greedy
    #            algorithm.

    #skymap    : a string that refers to the name of a txt file. 
    #            GW sky localisation information. The skymap should be a
    #            healpix file translated into a txt file that contains four columns:
    #           1. Index; 2. co-declination; 3. right ascension; 4. GW prob.
    #
    #resolution : a number that defines the resolution of the healpix file.
    #             The number of pixels of a skymap is equal to 12*2^(2*resolution)

    #level : a number that sets the confidence level of the credible region that 
    #        will be considered. The default value if 0.9. To use the default
    #       value, enter [].
 
    #PGW : a vector containing the values of the sum of the GW probability  
    #      within the fields

    #fields_location : a n by 2 matrix containing the location infomration of
    #                  the selected fields. The first coloumn corresponds to
    #                  the declination of the field centers, and the second
    #                  the right ascension.
    DelThe = np.sqrt(fov) * np.pi/180.0
    DelPhi = np.sqrt(fov) * np.pi/180.0
    
    searchrange = np.sqrt(4.0) * np.sqrt(fov) / 2 * np.pi / 180.0
    
    print('Loading the skymap information')
    
    #nos_file = ASDtxt(detector)
    index_str = []
    CoDec_str = []
    Ra_str = []
    GW_prob_str = []
    file = open(skymap, 'r')
    readFile = file.readlines()
    file.close()
    #f = []
    #ASD = []

    for line in readFile:
        p = line.split()
        index_str.append(float(p[0]))
        CoDec_str.append(float(p[1]))
        Ra_str.append(float(p[2]))
        GW_prob_str.append(float(p[3]))
        
    del readFile
    
    index = np.array(index_str)
    CoDec = np.array(CoDec_str)
    Ra = np.array(Ra_str)
    GW_prob = np.array(GW_prob_str)
    
    postinfo = np.array([index, CoDec, Ra, GW_prob])
    
    
    del index, CoDec, Ra, GW_prob
    
    sorted_postinfo = postinfo[:, postinfo[3,:].argsort()[::-1]]
    
    if not level:
        level = 0.9
    elif isinstance(level, numbers.Number) == False: 
        sys.exit('level has to either be an empty array or a number.')     
    
    post = sorted_postinfo[3,:]
    cumulativeSum = np.cumsum(post) / sum(post)
    
    indexthreshold = np.argmax(cumulativeSum >= level)
    
    message1='% Credible Region of the GW Posterior is being considered, and a sky map is being generated. \n';
    display_message=[str(level*100),message1];
    print(''.join(display_message))
    
    sorted_postinfo = sorted_postinfo[:, 0:indexthreshold]
    dec = np.pi / 2.0 - sorted_postinfo[1, :]
    ra = sorted_postinfo[2, :]
    
    ra[np.argwhere(ra > np.pi)] = ra[np.argwhere(ra > np.pi)] - 2.0 * np.pi
    colors = post
    get_ipython().magic(u'matplotlib qt')
    plot.subplot(111, projection='aitoff')
    skymap = plot.scatter(ra, dec, c = post[0:indexthreshold], s = 200)
    plot.colorbar(skymap)
    plot.show()

    
    Prob_in_field = np.zeros(n)
    
    x = np.cos(np.pi / 2.0 - postinfo[1,:]) * np.cos(postinfo[2,:])
    y = np.cos(np.pi / 2.0 - postinfo[1,:]) * np.sin(postinfo[2,:])
    z = np.sin(np.pi / 2.0 - postinfo[1,:])
    
    x1 = np.cos(np.pi / 2.0 - sorted_postinfo[1,:]) * np.cos(sorted_postinfo[2,:])
    y1 = np.cos(np.pi / 2.0 - sorted_postinfo[1,:]) * np.sin(sorted_postinfo[2,:])
    z1 = np.sin(np.pi / 2.0 - sorted_postinfo[1,:])    
    
    Cartesian = np.array([x,y,z])
    
    point_tree = spatial.cKDTree(np.transpose(Cartesian))
    idx = point_tree.query_ball_point(np.transpose(np.array([x1,y1,z1])), searchrange)
    
    num = 1
    PGW = np.zeros(n)
    
    fields_location = np.zeros((n,2))
    
    #margin = np.sqrt(4.0*np.pi/(12.0 * 2 ** (2.0*resolution)))
    
    #micro_movement1=np.array([0, 0, 0, -0.5, -0.5, -0.5, 0.5, 0.5, 0.5]);
    #micro_movement2=np.array([-0.5, 0, 0.5, -0.5, 0, 0.5, -0.5, 0, 0.5]);
    
    calibration = -0.3
    
    location = np.zeros((n, 6))
    resolution_degree = 0
    if resolution < 10:
        check = np.zeros(len(idx))
    while num <= n:
        if num % 10 == 1 and num != 11:
            print(''.join(['Locating the ', str(num), 'st field. Note this does not refer to the observing order.']))
        elif num % 10 == 2 and num != 12:
            print(''.join(['Locating the ', str(num), 'nd field. Note this does not refer to the observing order.']))
        elif num % 10 == 3 and num != 13:
            print(''.join(['Locating the ', str(num), 'rd field. Note this does not refer to the observing order.']))
        else:
            print(''.join(['Locating the ', str(num), 'th field. Note this does not refer to the observing order.']))
        
        P_fov = np.zeros(len(idx))
        filtered_idx = [[] for j in range(len(idx))]
        for i in range(len(idx)):
            phi = sorted_postinfo[1, i]
            theta = sorted_postinfo[2, i]

            filtered = postinfo[:, idx[i]]
            filtered_makezeros = postinfo[:, idx[i]]
            
            newFilter = Cartesian[:, filtered[0,:].astype(int)]
            newCartUp = np.dot(roty(phi - DelPhi/2.0 - np.pi/2.0), np.dot(rotz(theta), newFilter))
            newCartLo = np.dot(roty(phi + DelPhi/2.0 - np.pi/2.0), np.dot(rotz(theta), newFilter))
                
            newCartLeft = np.dot(rotz(DelThe/2.0), np.dot(rotx(np.pi/2 - phi), np.dot(rotz(theta - np.pi/2), newFilter)))
            
            newCartRight = np.dot(rotz(-DelThe/2.0), np.dot(rotx(np.pi/2 - phi), np.dot(rotz(theta - np.pi/2), newFilter)))
            
            
            
            Filter_idx = np.where(np.logical_and.reduce((newCartLeft[0,:]>=0, newCartRight[0,:]<=0,
                                                             newCartUp[2,:] <= 0, newCartLo[2,:] >=0)))
            
            filtered = filtered[:, Filter_idx].squeeze()
            check[i] = np.count_nonzero(filtered[3,:])
            
            P_fov[i] = sum(filtered[3,:])
            
            newCartUp_makezeros = np.dot(roty(phi - DelPhi/2.0 - resolution_degree - np.pi/2.0), np.dot(rotz(theta), newFilter))
            newCartLo_makezeros = np.dot(roty(phi + DelPhi/2.0 + resolution_degree - np.pi/2.0), np.dot(rotz(theta), newFilter))
                
            newCartLeft_makezeros = np.dot(rotz(DelThe/2.0 + resolution_degree), np.dot(rotx(np.pi/2 - phi), np.dot(rotz(theta - np.pi/2), newFilter)))
            newCartRight_makezeros = np.dot(rotz(-DelThe/2.0 - resolution_degree), np.dot(rotx(np.pi/2 - phi), np.dot(rotz(theta - np.pi/2), newFilter)))
    
            Filter_idx_makezeros = np.where(np.logical_and.reduce((newCartLeft_makezeros[0,:]>=0, newCartRight_makezeros[0,:]<=0,
                                                             newCartUp_makezeros[2,:] <= 0, newCartLo_makezeros[2,:] >=0)))
    
            filtered_makezeros = filtered_makezeros[:, Filter_idx_makezeros]
            filtered_idx[i] = filtered_makezeros[0,:]
            
            
        max_id = np.argwhere(P_fov == np.amax(P_fov)).squeeze().astype(int)
        postinfo[3, filtered_idx[max_id].astype(int)] = 0
        Prob_in_field[num-1] = P_fov[max_id]
        declowlim = np.pi / 2.0 - np.amax(postinfo[1, filtered_idx[max_id].astype(int).squeeze()])
        decuplim = np.pi / 2.0 - np.amin(postinfo[1, filtered_idx[max_id].astype(int).squeeze()])
        
        ra_within = postinfo[2, filtered_idx[max_id].astype(int)].squeeze();
        
        ra_within[np.argwhere(ra_within > np.pi)] = ra_within[np.argwhere(ra_within > np.pi)] - 2.0 * np.pi
                             
        ra_uplim = np.amax(ra_within)  
        ra_lowlim = np.amin(ra_within)
        
        ra_lo = sorted_postinfo[2, max_id]
        dec_lo = np.pi/2.0 - sorted_postinfo[1, max_id]
        
        
        location[num - 1, 0] = ra_lowlim
        location[num - 1, 1] = ra_uplim
        location[num - 1, 2] = declowlim
        location[num - 1, 3] = decuplim
        location[num - 1, 4] = ra_lo
        location[num - 1, 5] = dec_lo
         
        if ra_uplim - ra_lowlim < 300.0:
            plot.plot([ra_lowlim, ra_lowlim], [declowlim, decuplim], 'k', linewidth = 2.0)
            plot.plot([ra_lowlim, ra_uplim], [decuplim, decuplim], 'k', linewidth = 2.0)
            plot.plot([ra_uplim, ra_uplim], [declowlim, decuplim], 'k', linewidth = 2.0)
            plot.plot([ra_lowlim, ra_uplim], [declowlim, declowlim], 'k', linewidth = 2.0)
        else:
            plot.plot([ra_lowlim, ra_lowlim], [declowlim, decuplim], 'k', linewidth = 2.0)
            plot.plot([ra_uplim, ra_lowlim], [decuplim, decuplim], 'k', linewidth = 2.0)
            plot.plot([ra_uplim, ra_uplim], [declowlim, decuplim], 'k', linewidth = 2.0)
            plot.plot([ra_uplim, ra_lowlim], [declowlim, declowlim], 'k', linewidth = 2.0)
        
        plot.text(ra_lo, dec_lo, str(num))
        
        num += 1
        
        
            
        
    return location, check, Prob_in_field

def Solver(PGW, nof, tau, prob, T, method):
    #[time_allocated,individual_prob,probability]=Solver(PGW,nos,tau,prob,T)
    #Solver returns the detection probability, and the time allocations 

    #area_index is a vector containing the values of P_GW that is in.
    #descending order of the values of P_GW.

    #nof is the current number of fields being considered.
    #tau, prob are the data for P_EM.

    #T is the total observation time excluding the slew time or readout time 
    #or both.
    
    #time_allocated is the cell array that contains the time allocations.
    #For example, time_allocated{10} is the time allocation for the highest 
    #10 fields in terms of P_GW when only these 10 fields are being observed.

    #individual_prob is the detection prob. probability rendered by an
    #individual field.
    
    prob_a_field = np.zeros(nof)
    
    integral_val_fun = interpolate.splrep(tau, prob, w=1.0*np.ones(len(prob)), s=0)
    
    if method == 'Eq':
        #time_allocated = T/nof * np.ones(nof)
        #Test_solution = 
        time_allocation = T/nof * np.ones(nof)
    elif method == 'LM':
        Test_solution = np.ones(nof + 1)
        time_allocation = np.zeros(nof + 1)
        
        def core(Test_solution):
            integral_val_fun = interpolate.splrep(tau, prob, w=1.0*np.ones(len(prob)), s=0)
            num_var = len(Test_solution)
            dlambda = np.zeros(num_var)

            for i in range(num_var - 1):
                dlambda[i] = PGW[i] * interpolate.splev(Test_solution[i], integral_val_fun, der = 1, ext = 1) + Test_solution[-1]
            dlambda[-1] = sum(Test_solution[0:-1]) - T
            return dlambda
    
    
        time_allocation = optimize.root(core, Test_solution, method = 'lm', options = {'maxiter': 5500, 'ftol': 1e-16, 'xtol': 1e-16 })    
        time_allocation = time_allocation.x
        
    
    
    for i in range(nof):
        prob_a_field[i] = PGW[i] * interpolate.splev(time_allocation[i], integral_val_fun, der = 0, ext = 3)
    
    probability = sum(prob_a_field)           
        
    return time_allocation, prob_a_field, probability









#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import copy
import h5py
import numpy as np
import healpy as hp

from scipy.stats import norm
from scipy.special import gammaincinv

from astroquery.vizier import Vizier

from astropy.table import Table
from astropy.io import ascii
from astropy.cosmology import WMAP9 as cosmo
from astropy.table import Column
import astropy.units as u

# Unset row limits when querying Vizier
Vizier.ROW_LIMIT = -1


def get_catalog(params, map_struct):
    if not os.path.isdir(params["catalogDir"]):
        os.makedirs(params["catalogDir"])

    catalogFile = os.path.join(params["catalogDir"],
                               "%s.hdf5" % params["galaxy_catalog"])

    """AB Magnitude zero point."""
    MAB0 = -2.5 * np.log10(3631.e-23)
    pc_cm = 3.08568025e18
    const = 4. * np.pi * (10. * pc_cm)**2.

    if params["galaxy_catalog"] == "2MRS":
        if not os.path.isfile(catalogFile):
            import astropy.constants as c

            cat, = Vizier.get_catalogs('J/ApJS/199/26/table3')

            ra, dec = cat["RAJ2000"], cat["DEJ2000"]
            cz = cat["cz"]
            magk = cat["Ktmag"]

            z = (u.Quantity(cat['cz']) / c.c).to(u.dimensionless_unscaled)

            completeness = 0.5
            alpha = -1.0
            MK_star = -23.55
            MK_max = MK_star + 2.5 * np.log10(gammaincinv(alpha + 2,
                                                          completeness))
            MK = magk - cosmo.distmod(z)
            idx = (z > 0) & (MK < MK_max)

            ra, dec = ra[idx], dec[idx]
            z = z[idx]
            magk = magk[idx]

            distmpc = cosmo.luminosity_distance(z).to('Mpc').value

            with h5py.File(catalogFile, 'w') as f:
                f.create_dataset('ra', data=ra)
                f.create_dataset('dec', data=dec)
                f.create_dataset('z', data=z)
                f.create_dataset('magk', data=magk)
                f.create_dataset('distmpc', data=distmpc)

        else:
            with h5py.File(catalogFile, 'r') as f:
                ra, dec = f['ra'][:], f['dec'][:]
                z = f['z'][:]
                magk = f['magk'][:]
                distmpc = f['distmpc'][:]
        r = distmpc * 1.0
        mag = magk * 1.0

    elif params["galaxy_catalog"] == "GLADE":
        if not os.path.isfile(catalogFile):
            cat, = Vizier.get_catalogs('VII/281/glade2')

            ra, dec = cat["RAJ2000"], cat["DEJ2000"]
            distmpc, z = cat["Dist"], cat["z"]
            magb, magk = cat["Bmag"], cat["Kmag"]
            # Keep track of galaxy identifier
            GWGC, PGC, HyperLEDA = cat["GWGC"], cat["PGC"], cat["HyperLEDA"]
            _2MASS, SDSS = cat["_2MASS"], cat["SDSS-DR12"]

            idx = np.where(distmpc >= 0)[0]
            ra, dec = ra[idx], dec[idx]
            distmpc, z = distmpc[idx], z[idx]
            magb, magk = magb[idx], magk[idx]
            GWGC, PGC, HyperLEDA = GWGC[idx], PGC[idx], HyperLEDA[idx]
            _2MASS, SDSS = _2MASS[idx], SDSS[idx]

            with h5py.File(catalogFile, 'w') as f:
                f.create_dataset('ra', data=ra)
                f.create_dataset('dec', data=dec)
                f.create_dataset('distmpc', data=distmpc)
                f.create_dataset('magb', data=magb)
                f.create_dataset('magk', data=magk)
                f.create_dataset('z', data=z)
                # Add galaxy identifier
                f.create_dataset('GWGC', data=GWGC)
                f.create_dataset('PGC', data=PGC)
                f.create_dataset('HyperLEDA', data=HyperLEDA)
                f.create_dataset('2MASS', data=_2MASS)
                f.create_dataset('SDSS', data=SDSS)

        else:
            with h5py.File(catalogFile, 'r') as f:
                ra, dec = f['ra'][:], f['dec'][:]
                distmpc, z = f['distmpc'][:], f['z'][:]
                magb, magk = f['magb'][:], f['magk'][:]
                GWGC, PGC, _2MASS = f['GWGC'][:], f['PGC'][:], f['2MASS'][:]
                HyperLEDA, SDSS = f['HyperLEDA'][:], f['SDSS'][:]
                # Convert bytestring to unicode
                GWGC = GWGC.astype('U')
                PGC = PGC.astype('U')
                HyperLEDA = HyperLEDA.astype('U')
                _2MASS = _2MASS.astype('U')
                SDSS = SDSS.astype('U')

        # Keep only galaxies with finite B mag when using it in the grade
        if params["galaxy_grade"] == "S":
            idx = np.where(~np.isnan(magb))[0]
            ra, dec, distmpc = ra[idx], dec[idx], distmpc[idx]
            magb, magk = magb[idx], magk[idx]
            GWGC, PGC, HyperLEDA = GWGC[idx], PGC[idx], HyperLEDA[idx]
            _2MASS, SDSS = _2MASS[idx], SDSS[idx]

        r = distmpc * 1.0
        mag = magb * 1.0

    elif params["galaxy_catalog"] == "CLU":
        if not os.path.isfile(catalogFile):
            print("Please add %s." % catalogFile)
            exit(0)

        with h5py.File(catalogFile, 'r') as f:
            name = f['name'][:]
            ra, dec = f['ra'][:], f['dec'][:]
            sfr_fuv, mstar = f['sfr_fuv'][:], f['mstar'][:]
            distmpc, magb = f['distmpc'][:], f['magb'][:]
            a, b2a, pa = f['a'][:], f['b2a'][:], f['pa'][:]
            btc = f['btc'][:]

        idx = np.where(distmpc >= 0)[0]
        ra, dec = ra[idx], dec[idx]
        sfr_fuv, mstar = sfr_fuv[idx], mstar[idx]
        distmpc, magb = distmpc[idx], magb[idx]
        a, b2a, pa = a[idx], b2a[idx], pa[idx]
        btc = btc[idx]

        idx = np.where(~np.isnan(magb))[0]
        ra, dec = ra[idx], dec[idx]
        sfr_fuv, mstar = sfr_fuv[idx], mstar[idx]
        distmpc, magb = distmpc[idx], magb[idx]
        a, b2a, pa = a[idx], b2a[idx], pa[idx]
        btc = btc[idx]

        z = -1*np.ones(distmpc.shape)
        r = distmpc * 1.0
        mag = magb * 1.0

    L_nu = const * 10.**((mag + MAB0)/(-2.5))
    L_nu = np.log10(L_nu)
    L_nu = L_nu**params["catalog_n"]
    Slum = L_nu / np.sum(L_nu)

    mlim, M_KNmin, M_KNmax = 22, -17, -12
    L_KNmin = const * 10.**((M_KNmin + MAB0)/(-2.5))
    L_KNmax = const * 10.**((M_KNmax + MAB0)/(-2.5))

    Llim = 4. * np.pi * (r * 1e6 * pc_cm)**2. * 10.**((mlim + MAB0)/(-2.5))
    Sdet = (L_KNmax-Llim)/(L_KNmax-L_KNmin)
    Sdet[Sdet < 0.01] = 0.01
    Sdet[Sdet > 1.0] = 1.0

    n, cl = params["powerlaw_n"], params["powerlaw_cl"]
    dist_exp = params["powerlaw_dist_exp"]

    prob_scaled = copy.deepcopy(map_struct["prob"])
    prob_sorted = np.sort(prob_scaled)[::-1]
    prob_indexes = np.argsort(prob_scaled)[::-1]
    prob_cumsum = np.cumsum(prob_sorted)
    index = np.argmin(np.abs(prob_cumsum - cl)) + 1
    prob_scaled[prob_indexes[index:]] = 0.0
    prob_scaled = prob_scaled**n

    theta = 0.5 * np.pi - dec * 2 * np.pi / 360.0
    phi = ra * 2 * np.pi / 360.0
    ipix = hp.ang2pix(map_struct["nside"], ra, dec, lonlat=True)

    if "distnorm" in map_struct:
        if map_struct["distnorm"] is not None:
            Sloc = prob_scaled[ipix] * (map_struct["distnorm"][ipix] *
                                        norm(map_struct["distmu"][ipix],
                                        map_struct["distsigma"][ipix]).pdf(r))**params["powerlaw_dist_exp"] / map_struct["pixarea"]
        else:
            Sloc = copy.copy(prob_scaled[ipix])
    else:
        Sloc = copy.copy(prob_scaled[ipix])

    # Set nan values to zero
    Sloc[np.isnan(Sloc)] = 0

    S = Sloc*Slum*Sdet
    prob = np.zeros(map_struct["prob"].shape)
    if params["galaxy_grade"] == "Sloc":
        prob[ipix] = prob[ipix] + Sloc
        grade = Sloc
    elif params["galaxy_grade"] == "S":
        prob[ipix] = prob[ipix] + S
        grade = S

    prob = prob / np.sum(prob)

    map_struct['prob_catalog'] = prob
    if params["doUseCatalog"]:
        map_struct['prob'] = prob

    idx = np.where(~np.isnan(grade))[0]
    grade = grade[idx]
    ra, dec, Sloc, S = ra[idx], dec[idx], Sloc[idx], S[idx]
    distmpc, z = distmpc[idx], z[idx]
    if params["galaxy_catalog"] == "GLADE":
        GWGC, PGC, HyperLEDA = GWGC[idx], PGC[idx], HyperLEDA[idx]
        _2MASS, SDSS = _2MASS[idx], SDSS[idx]

    Sthresh = np.max(grade)*0.01
    idx = np.where(grade >= Sthresh)[0]
    grade = grade[idx]
    ra, dec, Sloc, S = ra[idx], dec[idx], Sloc[idx], S[idx]
    distmpc, z = distmpc[idx], z[idx]
    if params["galaxy_catalog"] == "GLADE":
        GWGC, PGC, HyperLEDA = GWGC[idx], PGC[idx], HyperLEDA[idx]
        _2MASS, SDSS = _2MASS[idx], SDSS[idx]

    idx = np.argsort(grade)[::-1]
    grade = grade[idx]
    ra, dec, Sloc, S = ra[idx], dec[idx], Sloc[idx], S[idx]
    distmpc, z = distmpc[idx], z[idx]
    if params["galaxy_catalog"] == "GLADE":
        GWGC, PGC, HyperLEDA = GWGC[idx], PGC[idx], HyperLEDA[idx]
        _2MASS, SDSS = _2MASS[idx], SDSS[idx]

    if len(ra) > 1000:
        print('Cutting catalog to top 1000 galaxies...')
        idx = np.arange(1000).astype(int)
        ra, dec, Sloc, S = ra[idx], dec[idx], Sloc[idx], S[idx]
        distmpc, z = distmpc[idx], z[idx]
        if params["galaxy_catalog"] == "GLADE":
            GWGC, PGC, HyperLEDA = GWGC[idx], PGC[idx], HyperLEDA[idx]
            _2MASS, SDSS = _2MASS[idx], SDSS[idx]

    catalog_struct = {}
    catalog_struct["ra"] = ra
    catalog_struct["dec"] = dec
    catalog_struct["Sloc"] = Sloc
    catalog_struct["S"] = S

    if params["writeCatalog"]:
        catalogfile = os.path.join(params["outputDir"], 'catalog.csv')
        fid = open(catalogfile, 'w')
        cnt = 1
        if params["galaxy_catalog"] == "GLADE":
            fid.write("id, RAJ2000, DEJ2000, Sloc, S, Dist, z, GWGC, PGC, HyperLEDA, 2MASS, SDSS\n")
            for a, b, c, d, e, f, g, h, i, j, k in zip(ra, dec, Sloc, S, distmpc, z, GWGC, PGC, HyperLEDA, _2MASS, SDSS):
                fid.write("%d, %.5f, %.5f, %.5e, %.5e, %.4f, %.4f, %s, %s, %s, %s, %s\n" % (cnt, a, b, c, d, e, f, g, h, i, j, k))
                cnt = cnt + 1
        else:
            fid.write("id, RAJ2000, DEJ2000, Sloc, S, Dist, z\n")
            for a, b, c, d in zip(ra, dec, Sloc, S, distmpc, z):
                fid.write("%d, %.5f, %.5f, %.5e, %.5e, %.4f, %.4f\n" % (cnt, a, b, c, d, e, f))
                cnt = cnt + 1

        fid.close()

    return map_struct, catalog_struct




import os, sys, optparse
import numpy as np
import healpy as hp
import json

from datetime import datetime
from astropy.time import Time

import matplotlib
#matplotlib.rc('text', usetex=True)
matplotlib.use('Agg')
matplotlib.rcParams.update({'font.size': 16})
matplotlib.rcParams['contour.negative_linestyle'] = 'solid'
import matplotlib.pyplot as plt

import ligo.gracedb
from ligo.gracedb.rest import GraceDb, HTTPError
from ligo.gracedb.rest import GraceDbBasic

url = 'https://gracedb-test.ligo.org/api/'
#url = 'https://gracedb-test.ligo.org/apibasic/'

def get_event(params):

    g = GraceDb()
    eventString = params["event"]
    events = g.events('%s'%eventString)
    event = [x for x in events][0]

    keys = ['graceid','gpstime','extra_attributes','group','links','created','far','instruments','labels','nevents','submitter','search','likelihood']
    fileorder = ['LALInference_skymap.fits.gz','bayestar.fits.gz','BW_skymap.fits','LIB_skymap.fits.gz','skyprobcc_cWB.fits']
    #fileorder = ['LALInference3d.fits.gz','bayestar3d.fits.gz','bayestar.fits.gz']

    fileorder = ['LALInference_skymap.fits.gz','bayestar.fits.gz','BW_skymap.fits','LIB_skymap.fits.gz','skyprobcc_cWB.fits']

    eventinfo = {}
    for key in keys:
        if not key in event: continue
        eventinfo[key] = event[key]
    eventinfo['gpstime'] = float(eventinfo['gpstime'])
    if eventinfo['far'] == None:
        eventinfo['far'] = np.nan

    triggerfile = "%s/%s.txt"%(params["outputDir"],eventinfo['graceid'])
    skymapfile = '%s/%s.fits'%(params["outputDir"],eventinfo['graceid'])
    #if os.path.isfile(triggerfile) and os.path.isfile(skymapfile):
    #    print("Already have info for %s... continuing."%event["graceid"])
    #    return
    print("Getting info for %s"%event["graceid"])

    mjds = [-1,-1]
    timediff = -1
    if 'CoincInspiral' in event['extra_attributes']:
        eventinfo['coinc'] = event['extra_attributes']['CoincInspiral']
    if 'SingleInspiral' in event['extra_attributes']:
        eventinfo['singles'] = {}
        for single in event['extra_attributes']['SingleInspiral']:
            eventinfo['singles'][single['ifo']] = single
            eventinfo['singles'][single['ifo']]['gpstime'] = single['end_time']+10**-9 * single['end_time_ns']

        if ("H1" in eventinfo['singles']) and ("L1" in eventinfo['singles']):
            eventinfo["H1_L1_difference"] = eventinfo['singles']['H1']["gpstime"] - eventinfo['singles']['L1']["gpstime"]
            t = Time([eventinfo['singles']['H1']["gpstime"],eventinfo['singles']['L1']["gpstime"]], format='gps', scale='utc')
            mjds = t.mjd
            timediff = eventinfo["H1_L1_difference"]

    if 'MultiBurst' in event['extra_attributes']:
        eventinfo['burst'] = event['extra_attributes']['MultiBurst']

        single_ifo_times = eventinfo['burst']['single_ifo_times'].split(",")           
        ifos = eventinfo['burst']['ifos'].split(",")

        if len(ifos) > 1 and len(single_ifo_times) > 1:
            ifo1 = ifos[0]
            gps1 = float(single_ifo_times[0])
    
            ifo2 = ifos[1]
            gps2 = float(single_ifo_times[1])
    
            eventinfo['burst'][ifo1] = {}
            eventinfo['burst'][ifo1]['gpstime'] = gps1

            eventinfo['burst'][ifo2] = {}
            eventinfo['burst'][ifo2]['gpstime'] = gps2
    
            if ("H1" in eventinfo['burst']) and ("L1" in eventinfo['burst']):
                eventinfo["H1_L1_difference"] = eventinfo['burst']['H1']["gpstime"] - eventinfo['burst']['L1']["gpstime"]
                t = Time([eventinfo['burst']['H1']["gpstime"],eventinfo['burst']['L1']["gpstime"]], format='gps', scale='utc')
                mjds = t.mjd
                timediff = eventinfo["H1_L1_difference"]
    
    try:       
        print("Looking for EM bright file...")
        r = g.files(eventinfo['graceid'], "Source_Classification_%s.json"%eventinfo['graceid'])
        emfile = open('embright.json','w')
        emfile.write(r.read())
        emfile.close()

        with open('embright.json') as data_file:    
            emdata = json.load(data_file)          

        os.system('rm embright.json')

        embright_keys = ["Prob remnant_mass_greater_than 0M_sun","Prob EMbright"]  
        ns_keys = ["Prob Mass2_less_than 3M_sun","Prob NS2"]

        embright_prob = -1
        for key in embright_keys:
            if not key in embright_keys: continue
            embright_prob = emdata[key]
            break
        ns_prob = -1
        for key in embright_keys: 
            if not key in embright_keys: continue
            ns_prob = emdata[key]
            break

        eventinfo['embright'] = {}
        eventinfo['embright']['embright'] = embright_prob
        eventinfo['embright']['ns'] = ns_prob

    except:
         print("No EM bright file...")            

    try:
        print("Looking for cWB file...")
        r = g.files(eventinfo['graceid'], "trigger_%.4f.txt"%eventinfo['gpstime']) 
        # r = g.files(eventinfo['graceid'], "eventDump.txt")
        cwbfile = open('trigger.txt','w')
        cwbfile.write(r.read())
        cwbfile.close()

        eventinfo['burst'] = {}
        lines = [line.rstrip('\n') for line in open('trigger.txt')]
        for line in lines:
            lineSplit = line.split(":")
            if len(lineSplit) < 2: continue
            key = lineSplit[0]
            value = filter(None, lineSplit[1].split(" "))
            eventinfo['burst'][lineSplit[0]] = value

        ifo1 = eventinfo['burst']['ifo'][0]
        gps1 = float(eventinfo['burst']['time'][0])

        ifo2 = eventinfo['burst']['ifo'][1]
        gps2 = float(eventinfo['burst']['time'][1])

        eventinfo['burst'][ifo1] = {}
        eventinfo['burst'][ifo1]['gpstime'] = gps1

        eventinfo['burst'][ifo2] = {}
        eventinfo['burst'][ifo2]['gpstime'] = gps2

        if ("H1" in eventinfo['burst']) and ("L1" in eventinfo['burst']):
            eventinfo["H1_L1_difference"] = eventinfo['burst']['H1']["gpstime"] - eventinfo['burst']['L1']["gpstime"]
            t = Time([eventinfo['burst']['H1']["gpstime"],eventinfo['burst']['L1']["gpstime"]], format='gps', scale='utc')
            mjds = t.mjd
            timediff = eventinfo["H1_L1_difference"]

    except:
        print("No cWB file...")

    ra = 0
    dec = 0

    r = []
    for lvfile in fileorder:
        #r = g.files(eventinfo['graceid'], lvfile)
        try:
            r = g.files(eventinfo['graceid'], lvfile)
            break
        except:
            continue
    if r == []:
        print("Download of skymaps file for %s failed..."%eventinfo['graceid'])
    else:
   
        skymap = open(skymapfile,'w')
        skymap.write(r.read())
        skymap.close()
      
    return skymapfile, eventinfo
 




import os, sys
import numpy as np
import healpy as hp
from astropy.time import Time
import matplotlib
#matplotlib.rc('text', usetex=True)
matplotlib.use('Agg')
matplotlib.rcParams.update({'font.size': 16})
matplotlib.rcParams['contour.negative_linestyle'] = 'solid'
import matplotlib.pyplot as plt

import gwemopt.utils

def compute_efficiency(params, map_struct, lightcurve_struct, coverage_struct):

    nside = params["nside"]
    Ninj = params["Ninj"]
    Ndet = params["Ndet"]
    gpstime = params["gpstime"]
    mjd_inj = Time(gpstime, format='gps', scale='utc').mjd
    #FOV_r = np.sqrt(float(params['FOV'])/np.pi)

    if params["doCatalog"]:
        prob_data_sorted = np.sort(map_struct["prob_catalog"])[::-1]
        prob_data_indexes = np.argsort(map_struct["prob_catalog"])[::-1]
    else:
        prob_data_sorted = np.sort(map_struct["prob"])[::-1]
        prob_data_indexes = np.argsort(map_struct["prob"])[::-1]
    prob_data_cumsum = np.cumsum(prob_data_sorted)
    rand_values = np.random.rand(Ninj,)
    dists = np.logspace(-1,3,1000)
    #dists = np.logspace(-1,np.log10(300),100)
    ndetections = np.zeros((len(dists),))

    ras = []
    decs = []

    for ii in range(Ninj):
        # this returns the index of the value in cumsum of probablity closest to a random value
        ipix = np.argmin(np.abs(prob_data_cumsum-rand_values[ii])) 
        # the point corresponding to that
        ra_inj = map_struct["ra"][prob_data_indexes][ipix]
        dec_inj = map_struct["dec"][prob_data_indexes][ipix]

        ras.append(ra_inj)
        decs.append(dec_inj)

        # THE SKY-LOCATION AS A HEALPIXEL ID
        pinpoint = hp.ang2pix(nside, theta=ra_inj, phi=dec_inj, lonlat=True)

        #idxs = np.where(np.sqrt((coverage_struct["data"][:,0]-ra_inj)**2 + (coverage_struct["data"][:,1]-dec_inj)**2) <= FOV_r)[0]
        #idxs_ra = np.where(np.abs(coverage_struct["data"][:,0]-ra_inj)<=coverage_struct["FOV"]/2.0)[0]
        #idxs_dec = np.where(np.abs(coverage_struct["data"][:,1]-dec_inj)<=coverage_struct["FOV"]/2.0)[0]
        #idxs = np.intersect1d(idxs_ra,idxs_dec)

        idxs = []
        for jj in range(len(coverage_struct["ipix"])):
            expPixels = coverage_struct["ipix"][jj]
            if pinpoint in expPixels:
                idxs.append(jj)
        if len(idxs) == 0:
            continue

        mjds = coverage_struct["data"][idxs,2]
        mags = coverage_struct["data"][idxs,3]
        filts = coverage_struct["filters"][idxs]

        detections = np.zeros((len(dists),))

        for mjd, mag, filt in zip(mjds,mags,filts):
            lightcurve_t = lightcurve_struct["t"] + mjd_inj
            lightcurve_mag = lightcurve_struct[filt]

            lightcurve_mag_interp = np.interp(mjd,lightcurve_t,lightcurve_mag)
            dist_threshold = (10**(((mag-lightcurve_mag_interp)/5.0)+1.0))/1e6

            idxs_detections = np.where(dists <= dist_threshold)[0]
            detections[idxs_detections] = detections[idxs_detections] + 1
        idxs_detections = np.where(detections >= Ndet)[0]
        ndetections[idxs_detections] = ndetections[idxs_detections] + 1
    efficiency = ndetections / Ninj
    efficiency_struct = {}
    efficiency_struct["ra"] = np.array(ras)
    efficiency_struct["dec"] = np.array(decs)
    efficiency_struct["efficiency"] = efficiency
    efficiency_struct["distances"] = dists
    save_efficiency_data(params, efficiency_struct, lightcurve_struct)
    efficiency_metric = calculate_efficiency_metric(params, efficiency_struct)
    save_efficiency_metric(params, os.path.join(params["outputDir"],"efficiency.txt"), efficiency_metric, lightcurve_struct)
    return efficiency_struct


def save_efficiency_data(params, efficiency_struct, lightcurve_struct):
    for i in range(0, len(efficiency_struct["distances"])):
        dist = efficiency_struct["distances"][i]
        eff = efficiency_struct["efficiency"][i]
        filename = os.path.join(params["outputDir"],'efficiency_' + lightcurve_struct['name'] + '.txt')
        if os.path.exists(filename):
            append_write = 'a'
            efficiency_file = open(filename, append_write)
        else:
            append_write = 'w'
            efficiency_file = open(filename, append_write)
            efficiency_file.write("Distance" + "\t" + "efficiency\n")
        efficiency_file.write(str(dist) + "\t" + str(eff) + "\n")
    efficiency_file.close()


def calculate_efficiency_metric(params, efficiency_struct):
    dist_sum = 0
    weighted_sum = 0
    for i in range(0, len(efficiency_struct["distances"])):
        dist = efficiency_struct["distances"][i]
        eff = efficiency_struct["efficiency"][i]
        dist_sum += dist * dist
        weighted_sum += dist * dist * eff
    metric = weighted_sum / dist_sum
    uncertainty = np.sqrt(metric * (1 - metric) / params["Ninj"])
    return (metric, uncertainty)


def save_efficiency_metric(params, efficiency_filename, efficiency_metric, lightcurve_struct):
    if os.path.exists(efficiency_filename):
        append_write = 'a'
        efficiency_file = open(efficiency_filename, append_write)
    else:
        append_write = 'w'
        efficiency_file = open(efficiency_filename, append_write)
        efficiency_file.write("tilesType\t" + "timeallocationType\t" + "scheduleType\t" + "Ntiles\t" + "efficiencyMetric\t" + "efficiencyMetric_err\t" + "injection\n")
    efficiency_file.write(params["tilesType"] + "\t" + params["timeallocationType"] + "\t" +  params["scheduleType"] + "\t" + str(params["Ntiles"]) + "\t" + str(efficiency_metric[0]) + "\t" + str(efficiency_metric[1]) + "\t" + lightcurve_struct["name"] + "\n")
    efficiency_file.close()




import os, sys
import copy
import numpy as np
import healpy as hp
import itertools

from scipy.stats import norm

import astropy.coordinates
from astropy.time import Time, TimeDelta
import astropy.units as u

import matplotlib
#matplotlib.rc('text', usetex=True)
matplotlib.use('Agg')
matplotlib.rcParams.update({'font.size': 16})
matplotlib.rcParams['contour.negative_linestyle'] = 'solid'
import matplotlib.pyplot as plt
import matplotlib.patches
import matplotlib.path

import ligo.segments as segments
import ligo.skymap.distance as ligodist


def readParamsFromFile(file):
    """@read gwemopt params file

    @param file
        gwemopt params file
    """

    params = {}
    if os.path.isfile(file):
        with open(file,'r') as f:
            for line in f:
                line_without_return = line.split("\n")
                line_split = line_without_return[0].split(" ")
                line_split = list(filter(None, line_split))
                if line_split:
                    try:
                        params[line_split[0]] = float(line_split[1])
                    except:
                        params[line_split[0]] = line_split[1]
    return params

def read_skymap(params,is3D=False,map_struct=None):

    if map_struct is None:
        map_struct = {}

        if params["doDatabase"]:
            models = params["models"]
            localizations_all = models.Localization.query.all()
            localizations = models.Localization.query.filter_by(dateobs=params["dateobs"],localization_name=params["localization_name"]).all()
            if localizations == None:
                print("No localization with dateobs=%s"%params["dateobs"])
                exit(0)
            else:
                prob_data = localizations[0].healpix
                prob_data = prob_data / np.sum(prob_data)
                map_struct["prob"] = prob_data
    
                distmu = localizations[0].distmu
                distsigma = localizations[0].distsigma
                distnorm = localizations[0].distnorm
    
                if distmu is None:            
                    map_struct["distmu"] = None
                    map_struct["distsigma"] = None
                    map_struct["distnorm"] = None
                else:
                    map_struct["distmu"] = np.array(distmu)
                    map_struct["distsigma"] = np.array(distsigma)
                    map_struct["distnorm"] = np.array(distnorm)
                    is3D = True
        else:
            filename = params["skymap"]
        
            if is3D:
                healpix_data = hp.read_map(filename, field=(0,1,2,3), verbose=False)
        
                distmu_data = healpix_data[1]
                distsigma_data = healpix_data[2]
                prob_data = healpix_data[0]
                norm_data = healpix_data[3]
        
                map_struct["distmu"] = distmu_data / params["DScale"]
                map_struct["distsigma"] = distsigma_data / params["DScale"]
                map_struct["prob"] = prob_data
                map_struct["distnorm"] = norm_data
    
            else:
                prob_data = hp.read_map(filename, field=0, verbose=False)
                prob_data = prob_data / np.sum(prob_data)
        
                map_struct["prob"] = prob_data

    natural_nside = hp.pixelfunc.get_nside(map_struct["prob"])
    nside = params["nside"]
    
    print("natural_nside =", natural_nside)
    print("nside =", nside)
    
    if not is3D:
        map_struct["prob"] = hp.ud_grade(map_struct["prob"],nside,power=-2)

    if is3D:
        if natural_nside != nside:
            map_struct["prob"], map_struct["distmu"],\
            map_struct["distsigma"], map_struct["distnorm"] = ligodist.ud_grade(map_struct["prob"],\
                                                                                map_struct["distmu"],\
                                                                                map_struct["distsigma"],\
                                                                                nside)
        
        nside_down = 32
        _, distmu_down,\
        distsigma_down, distnorm_down = ligodist.ud_grade(map_struct["prob"],\
                                                          map_struct["distmu"],\
                                                          map_struct["distsigma"],\
                                                          nside_down)

        r = np.linspace(0, 2000)
        map_struct["distmed"] = np.zeros(distmu_down.shape)
        for ipix in range(len(map_struct["distmed"])):
            dp_dr = r**2 * distnorm_down[ipix] * norm(distmu_down[ipix],distsigma_down[ipix]).pdf(r)
            dp_dr_norm = np.cumsum(dp_dr / np.sum(dp_dr))
            idx = np.argmin(np.abs(dp_dr_norm-0.5))
            map_struct["distmed"][ipix] = r[idx]
        map_struct["distmed"] = hp.ud_grade(map_struct["distmu"],nside,power=-2)

    npix = hp.nside2npix(nside)
    theta, phi = hp.pix2ang(nside, np.arange(npix))
    ra = np.rad2deg(phi)
    dec = np.rad2deg(0.5*np.pi - theta)

    map_struct["ra"] = ra
    map_struct["dec"] = dec

    sort_idx = np.argsort(map_struct["prob"])[::-1]
    csm = np.empty(len(map_struct["prob"]))
    csm[sort_idx] = np.cumsum(map_struct["prob"][sort_idx])

    map_struct["cumprob"] = csm

    pixarea = hp.nside2pixarea(nside)
    pixarea_deg2 = hp.nside2pixarea(nside, degrees=True)

    map_struct["nside"] = nside
    map_struct["npix"] = npix
    map_struct["pixarea"] = pixarea
    map_struct["pixarea_deg2"] = pixarea_deg2

    return map_struct   

def samples_from_skymap(map_struct, is3D = False, Nsamples = 100):

    prob_data_sorted = np.sort(map_struct["prob"])[::-1]
    prob_data_indexes = np.argsort(map_struct["prob"])[::-1]
    prob_data_cumsum = np.cumsum(prob_data_sorted)

    rand_values = np.random.rand(Nsamples,)

    ras = []
    decs = []
    dists = []

    if is3D:
        r = np.linspace(0, 2000)
        rand_values_dist = np.random.rand(Nsamples,)

    for ii in range(Nsamples):
        ipix = np.argmin(np.abs(prob_data_cumsum-rand_values[ii]))
        ra_inj = map_struct["ra"][prob_data_indexes][ipix]
        dec_inj = map_struct["dec"][prob_data_indexes][ipix]

        ras.append(ra_inj)
        decs.append(dec_inj)    

        if is3D:
            dp_dr = r**2 * map_struct["distnorm"][prob_data_indexes][ipix] * norm(map_struct["distmu"][prob_data_indexes][ipix], map_struct["distsigma"][prob_data_indexes][ipix]).pdf(r)
            dp_dr_norm = np.cumsum(dp_dr / np.sum(dp_dr))
            idx = np.argmin(np.abs(dp_dr_norm-rand_values_dist[ii]))
            dist_inj = r[idx]
            dists.append(dist_inj)
        else:
            dists.append(50.0)

    samples_struct = {}
    samples_struct["ra"] = np.array(ras)
    samples_struct["dec"] = np.array(decs)
    samples_struct["dist"] = np.array(dists)

    return samples_struct

def get_ellipse_coords(a=0.0, b=0.0, x=0.0, y=0.0, angle=0.0, npts=10):
    """ Draws an ellipse using (360*k + 1) discrete points; based on pseudo code
    given at http://en.wikipedia.org/wiki/Ellipse
    k = 1 means 361 points (degree by degree)
    a = major axis distance,
    b = minor axis distance,
    x = offset along the x-axis
    y = offset along the y-axis
    angle = clockwise rotation [in degrees] of the ellipse;
        * angle=0  : the ellipse is aligned with the positive x-axis
        * angle=30 : rotated 30 degrees clockwise from positive x-axis
    """
    pts = np.zeros((npts, 2))

    beta = -angle * np.pi/180.0
    sin_beta = np.sin(beta)
    cos_beta = np.cos(beta)
    alpha = np.linspace(0,2*np.pi,npts) 

    sin_alpha = np.sin(alpha)
    cos_alpha = np.cos(alpha)
    
    pts[:, 0] = x + (a * cos_alpha * cos_beta - b * sin_alpha * sin_beta)
    pts[:, 1] = y + (a * cos_alpha * sin_beta + b * sin_alpha * cos_beta)

    return pts

def getCirclePixels(ra_pointing, dec_pointing, radius, nside, alpha=0.4, color='#FFFFFF', edgecolor='#FFFFFF'):

    theta = 0.5 * np.pi - np.deg2rad(dec_pointing)
    phi = np.deg2rad(ra_pointing)

    xyz = hp.ang2vec(theta, phi)
    ipix = hp.query_disc(nside, xyz, np.deg2rad(radius))

    radecs = get_ellipse_coords(a=radius/np.cos(np.deg2rad(dec_pointing)), b=radius, x=ra_pointing, y=dec_pointing, angle=0.0, npts=25)
    idx = np.where(radecs[:,1] > 90.0)[0]
    radecs[idx,1] = 180.0 - radecs[idx,1] 
    idx = np.where(radecs[:,1] < -90.0)[0]
    radecs[idx,1] = -180.0 - radecs[idx,1]    
    idx = np.where(radecs[:,0] > 360.0)[0]
    radecs[idx,0] = 720.0 - radecs[idx,0]
    idx = np.where(radecs[:,0] < 0.0)[0]
    radecs[idx,0] = 360.0 + radecs[idx,0]

    radecs = np.array(radecs)
    idx1 = np.where(radecs[:,0]>=180.0)[0]
    idx2 = np.where(radecs[:,0]<180.0)[0]
    idx3 = np.where(radecs[:,0]>300.0)[0]
    idx4 = np.where(radecs[:,0]<60.0)[0]
    if (len(idx1)>0 and len(idx2)>0) and not (len(idx3)>0 and len(idx4)>0):
        alpha = 0.0

    xyz = hp.ang2vec(radecs[:,0],radecs[:,1],lonlat=True)

    proj = hp.projector.MollweideProj(rot=None, coord=None)
    x,y = proj.vec2xy(xyz[:,0],xyz[:,1],xyz[:,2])
    xy = np.zeros(radecs.shape)
    xy[:,0] = x
    xy[:,1] = y
    #path = matplotlib.path.Path(xyz[:,1:3])
    path = matplotlib.path.Path(xy)
    patch = matplotlib.patches.PathPatch(path, alpha=alpha, color=color, fill=True, zorder=3, edgecolor=edgecolor)

    area = np.pi * radius**2

    return ipix, radecs, patch, area

def getSquarePixels(ra_pointing, dec_pointing, tileSide, nside, alpha = 0.4, color='#FFFFFF', edgecolor='#FFFFFF'):

    area = tileSide*tileSide

    decCorners = (dec_pointing - tileSide / 2.0, dec_pointing + tileSide / 2.0)

    #security for the periodic limit conditions 
    radecs = []
    for d in decCorners:
        if d > 90.:
            d = 180. - d
        elif d < -90.:
            d = -180 - d

        raCorners = (ra_pointing - (tileSide / 2.0) / np.cos(np.deg2rad(d)) , ra_pointing + (tileSide / 2.0) / np.cos(np.deg2rad(d)))

        #security for the periodic limit conditions 
        for r in raCorners:
            if r > 360.:
                r = r - 360.
            elif r < 0.:
                r = 360. + r
            radecs.append([r,d])

    radecs = np.array(radecs)
    idx1 = np.where(radecs[:,0]>=180.0)[0] 
    idx2 = np.where(radecs[:,0]<180.0)[0]
    idx3 = np.where(radecs[:,0]>300.0)[0]
    idx4 = np.where(radecs[:,0]<60.0)[0]
    if (len(idx1)>0 and len(idx2)>0) and not (len(idx3)>0 and len(idx4)>0):
        alpha = 0.0

    idx1 = np.where(np.abs(radecs[:,1])>=87.0)[0] 
    if len(idx1) == 4:
        return [], [], [], []

    idx1 = np.where((radecs[:,1]>=87.0) | (radecs[:,1]<=-87.0))[0]
    if len(idx1)>0:
        radecs = np.delete(radecs, idx1[0], 0)

    xyz = []
    for r, d in radecs:
        xyz.append(hp.ang2vec(r, d, lonlat=True))

    npts, junk = radecs.shape
    if npts == 4:
        xyz = [xyz[0], xyz[1],xyz[3], xyz[2]]
        ipix = hp.query_polygon(nside, np.array(xyz))
    else:
        ipix = hp.query_polygon(nside, np.array(xyz))

    #idx1 = np.where((radecs[:,1]>=70.0) | (radecs[:,1]<=-70.0))[0]
    #idx2 = np.where((radecs[:,0]>300.0) | (radecs[:,0]<60.0))[0]
    #if (len(idx1) == 0) or (len(idx2) > 0):
    #    return [], [], [], []

    xyz = np.array(xyz)
    proj = hp.projector.MollweideProj(rot=None, coord=None) 
    x,y = proj.vec2xy(xyz[:,0],xyz[:,1],xyz[:,2])
    xy = np.zeros(radecs.shape)
    xy[:,0] = x
    xy[:,1] = y
    path = matplotlib.path.Path(xy)
    patch = matplotlib.patches.PathPatch(path, alpha=alpha, color=color, fill=True, zorder=3, edgecolor=edgecolor)
    
    return ipix, radecs, patch, area

def integrationTime(T_obs, pValTiles, func=None, T_int=60.0):
    '''
    METHOD :: This method accepts the probability values of the ranked tiles, the 
              total observation time and the rank of the source tile. It returns 
              the array of time to be spent in each tile which is determined based
              on the localizaton probability of the tile. How the weight factor is 
              computed can also be supplied in functional form. Default is linear.
                      
    pValTiles :: The probability value of the ranked tiles. Obtained from ZTF_RT 
                             output
    T_obs     :: Total observation time available for the follow-up.
    func      :: functional form of the weight. Default is linear. 
                             For example, use x**2 to use a quadratic function.
    '''

    if func is None:
            f = lambda x: x
    else:
            f = lambda x: eval(func)
    fpValTiles = f(pValTiles)
    modified_prob = fpValTiles/np.sum(fpValTiles)
    modified_prob[np.isnan(modified_prob)] = 0.0
    t_tiles = modified_prob * T_obs ### Time spent in each tile if not constrained
    #t_tiles[t_tiles > 1200.0] = 1200.0 ### Upper limit of exposure time
    #t_tiles[t_tiles < 60] = 60.0 ### Lower limit of exposure time
    t_tiles = T_int*np.round(t_tiles/T_int)
    #Obs = np.cumsum(t_tiles) <= T_obs ### Tiles observable in T_obs seconds
    #time_per_tile = t_tiles[Obs] ### Actual time spent per tile

    return t_tiles

def observability(params, map_struct):

    airmass = params["airmass"]
    nside = params["nside"]
    npix = hp.nside2npix(nside)
    gpstime = params["gpstime"]
    event_time = Time(gpstime, format='gps', scale='utc')
    dts = np.arange(0,7,1.0/24.0)
    dts = np.arange(0,7,1.0/4.0)

    observatory_struct = {}

    for telescope in params["telescopes"]:
        config_struct = params["config"][telescope]

        observatory = astropy.coordinates.EarthLocation(
            lat=config_struct["latitude"]*u.deg, lon=config_struct["longitude"]*u.deg, height=config_struct["elevation"]*u.m)

        # Look up (celestial) spherical polar coordinates of HEALPix grid.
        theta, phi = hp.pix2ang(nside, np.arange(npix))
        # Convert to RA, Dec.
        radecs = astropy.coordinates.SkyCoord(
            ra=phi*u.rad, dec=(0.5*np.pi - theta)*u.rad)

        observatory_struct[telescope] = {}
        observatory_struct[telescope]["prob"] = copy.deepcopy(map_struct["prob"])
        observatory_struct[telescope]["observability"] = np.zeros((npix,))
        observatory_struct[telescope]["dts"] = {}

        for dt in dts:
            time = event_time+TimeDelta(dt*u.day)

            # Alt/az reference frame at observatory, now
            frame = astropy.coordinates.AltAz(obstime=time, location=observatory)
            # Transform grid to alt/az coordinates at observatory, now
            altaz = radecs.transform_to(frame)

            # Where is the sun, now?
            sun_altaz = astropy.coordinates.get_sun(time).transform_to(altaz)

            # How likely is it that the (true, unknown) location of the source
            # is within the area that is visible, now? Demand that sun is at
            # least 18 degrees below the horizon and that the airmass
            # (secant of zenith angle approximation) is at most 2.5.
            idx = np.where((altaz.alt >= 30*u.deg) &  (sun_altaz.alt <= -18*u.deg) & (altaz.secz <= airmass))[0]
            observatory_struct[telescope]["dts"][dt] = np.zeros((npix,))
            observatory_struct[telescope]["dts"][dt][idx] = 1
            observatory_struct[telescope]["observability"][idx] = 1
        observatory_struct[telescope]["prob"] = observatory_struct[telescope]["prob"]*observatory_struct[telescope]["observability"]

    return observatory_struct

def get_exposures(params, config_struct, segmentlist):
    '''
    Convert the availability times to a list segments with the length of telescope exposures.
    segmentlist: the segments that the telescope can do the follow-up.
    '''
    exposurelist = segments.segmentlist()

    exposure_time = np.max(params["exposuretimes"])

    for ii in range(len(segmentlist)):
        start_segment, end_segment = segmentlist[ii][0], segmentlist[ii][1]
        exposures = np.arange(start_segment, end_segment, exposure_time/86400.0)
        #exposurelist = np.append(exposurelist,exposures)

        for jj in range(len(exposures)):
            exposurelist.append(segments.segment(exposures[jj],exposures[jj]+exposure_time/86400.0))

    return exposurelist






import os
import numpy as np
import healpy as hp

def get_skymap(params):
   
    nside = params["nside"]

    npix = hp.nside2npix(nside)
    theta, phi = hp.pix2ang(nside, np.arange(npix))
    ra = np.rad2deg(phi)
    dec = np.rad2deg(0.5*np.pi - theta)

    footprint_ra, footprint_dec, footprint_radius = params["footprint_ra"], params["footprint_dec"], params["footprint_radius"]

    ra1, d1 = np.deg2rad(ra), np.deg2rad(dec)
    ra2, d2 = np.deg2rad(footprint_ra), np.deg2rad(footprint_dec)

    # Calculate angle between target and moon
    cosA = np.sin(d1)*np.sin(d2) + np.cos(d1)*np.cos(d2)*np.cos(ra1-ra2)
    angle = np.arccos(cosA)*(360/(2*np.pi))
    prob = np.exp(-np.power(angle, 2.) / (2 * np.power(footprint_radius, 2.)))
    prob = prob / np.sum(prob)

    fitsfile = os.path.join(params["outputDir"],'skymap.fits')
    hp.fitsfunc.write_map(fitsfile,prob)

    return fitsfile




import os, sys, optparse
import numpy as np
import healpy as hp
import json

from datetime import datetime
from astropy.time import Time

import matplotlib
#matplotlib.rc('text', usetex=True)
matplotlib.use('Agg')
matplotlib.rcParams.update({'font.size': 16})
matplotlib.rcParams['contour.negative_linestyle'] = 'solid'
import matplotlib.pyplot as plt

from scipy.stats import norm

from ligo.gracedb.rest import GraceDb, HTTPError
from ligo.gracedb.rest import GraceDbBasic

url = 'https://gracedb-test.ligo.org/api/'
#url = 'https://gracedb-test.ligo.org/apibasic/'

def parse_commandline():
    """
    Parse the options given on the command-line.
    """
    parser = optparse.OptionParser()

    parser.add_option("--doGPSLoop",  action="store_true", default=False)
    parser.add_option("--doGPS",  action="store_true", default=False)
    parser.add_option("--doFAR",  action="store_true", default=False)
    parser.add_option("--doType",  action="store_true", default=False)
    parser.add_option("--doLabel",  action="store_true", default=False)
    parser.add_option("--doEvent",  action="store_true", default=False)

    parser.add_option("-f", "--far", help="far threshold",type=float,default=1e-7)
    parser.add_option("-t", "--type", help="trigger type",default="LowMass")
    parser.add_option("-l", "--label", help="trigger label",default="EM_READY & ADVOK & ~INJ")
    parser.add_option("-s", "--startGPS", help="start gps",type=float,default=1135136300)
    parser.add_option("-e", "--endGPS", help="end gps",type=float,default=1135136400)    
    parser.add_option("-o", "--outputDir", help="output directory",default="output")
    #parser.add_option("-r", "--ra", help="ra (deg)",type=float,default=138.30788)
    #parser.add_option("-d", "--dec", help="dec (deg)",type=float,default=61.09263)
    #parser.add_option("-z", "--dist", help="distance (mpc)",type=float,default=990.0)

    parser.add_option("-r", "--ra", help="ra (deg)",type=float,default=122.9671)
    parser.add_option("-d", "--dec", help="dec (deg)",type=float,default=25.4226)
    parser.add_option("-z", "--dist", help="distance (mpc)",type=float,default=982.2)

    parser.add_option("-n", "--event", help="event name",default="G268556")


    opts, args = parser.parse_args()

    return opts

def get_eventstring(opts):
    eventString = ''
    if opts.doEvent:
        eventString = opts.event
        return eventString

    if opts.doGPS:
        eventString = '%s %.1f .. %.1f'%(eventString,opts.startGPS,opts.endGPS)
    if opts.doLabel:
        eventString = '%s %s'%(eventString,opts.label)
    if opts.doType:
        eventString = '%s %s'%(eventString,opts.type)
    if opts.doFAR:
        eventString = '%s far <%.5e'%(eventString,opts.far)
    return eventString

def download_events(events):

    keys = ['graceid','gpstime','extra_attributes','group','links','created','far','instruments','labels','nevents','submitter','search','likelihood']
    fileorder = ['LALInference_skymap.fits.gz','bayestar.fits.gz','BW_skymap.fits','LIB_skymap.fits.gz','skyprobcc_cWB.fits']
    #fileorder = ['LALInference3d.fits.gz','bayestar3d.fits.gz','bayestar.fits.gz']

    fileorder = ['LALInference_skymap.fits.gz','bayestar.fits.gz','BW_skymap.fits','LIB_skymap.fits.gz','skyprobcc_cWB.fits']

    for event in events:
        eventinfo = {}
        for key in keys:
            if not key in event: continue
            eventinfo[key] = event[key]
        eventinfo['gpstime'] = float(eventinfo['gpstime'])
        if eventinfo['far'] == None:
            eventinfo['far'] = np.nan

        #for key in event:
        #    print key
 
        #if not eventinfo['graceid'] == "G211117": continue
 
        triggerfile = "%s/%s.txt"%(triggerDir,eventinfo['graceid'])
        #skymapfile = '%s/%s.fits.gz'%(skymapDir,eventinfo['graceid'])
        skymapfile = '%s/%s.fits'%(skymapDir,eventinfo['graceid'])
        if os.path.isfile(triggerfile) and os.path.isfile(skymapfile):
            print "Already have info for %s... continuing."%event["graceid"]
            continue
        print "Getting info for %s"%event["graceid"]

        mjds = [-1,-1]
        timediff = -1
        if 'CoincInspiral' in event['extra_attributes']:
            eventinfo['coinc'] = event['extra_attributes']['CoincInspiral']
        if 'SingleInspiral' in event['extra_attributes']:
            eventinfo['singles'] = {}
            for single in event['extra_attributes']['SingleInspiral']:
                eventinfo['singles'][single['ifo']] = single
                eventinfo['singles'][single['ifo']]['gpstime'] = single['end_time']+10**-9 * single['end_time_ns']
    
            if ("H1" in eventinfo['singles']) and ("L1" in eventinfo['singles']):
                eventinfo["H1_L1_difference"] = eventinfo['singles']['H1']["gpstime"] - eventinfo['singles']['L1']["gpstime"]
                t = Time([eventinfo['singles']['H1']["gpstime"],eventinfo['singles']['L1']["gpstime"]], format='gps', scale='utc')
                mjds = t.mjd
                timediff = eventinfo["H1_L1_difference"]
    
        if 'MultiBurst' in event['extra_attributes']:
            eventinfo['burst'] = event['extra_attributes']['MultiBurst']

            single_ifo_times = eventinfo['burst']['single_ifo_times'].split(",")           
            ifos = eventinfo['burst']['ifos'].split(",")

            if len(ifos) > 1 and len(single_ifo_times) > 1:
                ifo1 = ifos[0]
                gps1 = float(single_ifo_times[0])
        
                ifo2 = ifos[1]
                gps2 = float(single_ifo_times[1])
        
                eventinfo['burst'][ifo1] = {}
                eventinfo['burst'][ifo1]['gpstime'] = gps1
    
                eventinfo['burst'][ifo2] = {}
                eventinfo['burst'][ifo2]['gpstime'] = gps2
        
                if ("H1" in eventinfo['burst']) and ("L1" in eventinfo['burst']):
                    eventinfo["H1_L1_difference"] = eventinfo['burst']['H1']["gpstime"] - eventinfo['burst']['L1']["gpstime"]
                    t = Time([eventinfo['burst']['H1']["gpstime"],eventinfo['burst']['L1']["gpstime"]], format='gps', scale='utc')
                    mjds = t.mjd
                    timediff = eventinfo["H1_L1_difference"]
        
        try:       
            print "Looking for EM bright file..."
            r = g.files(eventinfo['graceid'], "Source_Classification_%s.json"%eventinfo['graceid'])
            emfile = open('embright.json','w')
            emfile.write(r.read())
            emfile.close()

            with open('embright.json') as data_file:    
                emdata = json.load(data_file)          

            os.system('rm embright.json')

            embright_keys = ["Prob remnant_mass_greater_than 0M_sun","Prob EMbright"]  
            ns_keys = ["Prob Mass2_less_than 3M_sun","Prob NS2"]

            embright_prob = -1
            for key in embright_keys:
                if not key in embright_keys: continue
                embright_prob = emdata[key]
                break
            ns_prob = -1
            for key in embright_keys: 
                if not key in embright_keys: continue
                ns_prob = emdata[key]
                break

            eventinfo['embright'] = {}
            eventinfo['embright']['embright'] = embright_prob
            eventinfo['embright']['ns'] = ns_prob

        except:
             print "No EM bright file..."            

        try:
            print "Looking for cWB file..."
            r = g.files(eventinfo['graceid'], "trigger_%.4f.txt"%eventinfo['gpstime']) 
            # r = g.files(eventinfo['graceid'], "eventDump.txt")
            cwbfile = open('trigger.txt','w')
            cwbfile.write(r.read())
            cwbfile.close()
    
            eventinfo['burst'] = {}
            lines = [line.rstrip('\n') for line in open('trigger.txt')]
            for line in lines:
                lineSplit = line.split(":")
                if len(lineSplit) < 2: continue
                key = lineSplit[0]
                value = filter(None, lineSplit[1].split(" "))
                eventinfo['burst'][lineSplit[0]] = value
    
            ifo1 = eventinfo['burst']['ifo'][0]
            gps1 = float(eventinfo['burst']['time'][0])
    
            ifo2 = eventinfo['burst']['ifo'][1]
            gps2 = float(eventinfo['burst']['time'][1])
    
            eventinfo['burst'][ifo1] = {}
            eventinfo['burst'][ifo1]['gpstime'] = gps1
    
            eventinfo['burst'][ifo2] = {}
            eventinfo['burst'][ifo2]['gpstime'] = gps2
    
            if ("H1" in eventinfo['burst']) and ("L1" in eventinfo['burst']):
                eventinfo["H1_L1_difference"] = eventinfo['burst']['H1']["gpstime"] - eventinfo['burst']['L1']["gpstime"]
                t = Time([eventinfo['burst']['H1']["gpstime"],eventinfo['burst']['L1']["gpstime"]], format='gps', scale='utc')
                mjds = t.mjd
                timediff = eventinfo["H1_L1_difference"]
    
        except:
            print "No cWB file..."
    
        #if len(gpss) > 0:
        #    dist = np.absolute(eventinfo['gpstime'] - np.array(gpss))
        #    if np.min(dist) < 5:
        #        continue
        #gpss.append(eventinfo['gpstime'])
    
        ra = 0
        dec = 0
    
        r = []
        for lvfile in fileorder:
            #r = g.files(eventinfo['graceid'], lvfile)
            try:
                r = g.files(eventinfo['graceid'], lvfile)
                break
            except:
                continue
        if r == []:
            print "Download of skymaps file for %s failed..."%eventinfo['graceid']
        else:
   
            skymap = open(skymapfile,'w')
            skymap.write(r.read())
            skymap.close()
       
            print skymapfile 
            prob = hp.read_map(skymapfile)
            try:
                prob, distmu, distsigma, distnorm = hp.read_map(skymapfile,field=(0,1,2,3)) 
                distanceInfo = True
            except:
                distanceInfo = False
                print "No distance information"
            ra, dec, dist = opts.ra, opts.dec, opts.dist
            theta = 0.5 * np.pi - np.deg2rad(dec)
            phi = np.deg2rad(ra)

            npix = len(prob)
            nside = hp.npix2nside(npix)

            if distanceInfo:
                ipix = hp.ang2pix(nside, theta, phi)
                pixarea = hp.nside2pixarea(nside)
                pixarea_deg2 = hp.nside2pixarea(nside, degrees=True)

                dp_dA = prob[ipix] / pixarea
                dp_dA_deg2 = prob[ipix] / pixarea_deg2
     
                r = np.linspace(0, 1500)
                dp_dr = r**2 * distnorm[ipix] * norm(distmu[ipix], distsigma[ipix]).pdf(r)
                dp_dV = prob[ipix] * distnorm[ipix] * norm(distmu[ipix], distsigma[ipix]).pdf(dist) / pixarea
                dp_dr_sky = [np.sum(prob*rr**2 * distnorm * norm(distmu, distsigma).pdf(rr)) for rr in r]
                dp_dr_norm = np.cumsum(dp_dr / np.sum(dp_dr))
                dp_dr_sky_norm = np.cumsum(dp_dr_sky / np.sum(dp_dr_sky))
 
                dp_dr_norm_10 = np.argmin(np.abs(dp_dr_norm - 0.1))
                dp_dr_norm_90 = np.argmin(np.abs(dp_dr_norm - 0.9))
                dp_dr_sky_norm_10 = np.argmin(np.abs(dp_dr_sky_norm - 0.1))
                dp_dr_sky_norm_90 = np.argmin(np.abs(dp_dr_sky_norm - 0.9))

                print "Probability per steradian: %.5e"%dp_dA 
                print "Probability per deg^2: %.5e"%dp_dA_deg2
                print "Probability per unit volume: %.5e / Mpc^3"%dp_dV
                print "Line-of-sight 10-90 (mpc): %.0f - %.0f"%(r[dp_dr_norm_10],r[dp_dr_norm_90])
                print "All-sky 10-90 (mpc): %.0f - %.0f"%(r[dp_dr_sky_norm_10],r[dp_dr_sky_norm_90])

            index = np.argmax(prob)
            theta, phi = hp.pix2ang(nside, index)
            ra = phi
            dec = 0.5*np.pi - theta
        
            ra = (ra / (2*np.pi)) * 24.0
            dec = dec * (360.0/(2*np.pi))
      
            thisplotDir = os.path.join(plotDir,eventinfo['graceid'])
            if not os.path.isdir(thisplotDir):
                os.mkdir(thisplotDir)
        
            plotName = os.path.join(thisplotDir,'mollview.png')
            hp.mollview(prob)
            plt.show()
            plt.savefig(plotName,dpi=200)
            plt.close('all')
   
            if distanceInfo:
                plotName = os.path.join(thisplotDir,'prob_dist.png')
                plt.figure(figsize=(8, 6))
                plt.plot(r, dp_dr)
                plt.xlabel('distance (Mpc)')
                plt.ylabel('prob Mpc$^{-1}$')
                plt.show()
                plt.savefig(plotName,dpi=200)
                plt.close('all')
 
                plotName = os.path.join(thisplotDir,'prob_dist_sky.png')
                plt.figure(figsize=(8, 6))
                plt.plot(r, dp_dr_sky)
                plt.plot([dist,dist],[0,np.max(dp_dr_sky)],'r--')
                plt.xlabel('distance (Mpc)')
                plt.ylabel('prob Mpc$^{-1}$')
                plt.show()
                plt.savefig(plotName,dpi=200)
                plt.close('all')
 
        f = open(triggerfile,"w")
        print "%s %.0f %.10f %.10f %.10f %.10f %.10f %.10f"%(eventinfo['graceid'], eventinfo['gpstime'], eventinfo['far'], ra, dec, mjds[0], mjds[1],timediff)
        f.write("%.0f %.10f %.10f %.10f %.10f %.10f\n"%(eventinfo['gpstime'],ra,dec,mjds[0],mjds[1],timediff))
        f.close()

opts = parse_commandline()
 
triggerDir = os.path.join(opts.outputDir,"triggers")
skymapDir = os.path.join(opts.outputDir,"skymaps")
plotDir = os.path.join(opts.outputDir,"plots")
 
if not os.path.isdir(opts.outputDir): os.mkdir(opts.outputDir)
if not os.path.isdir(triggerDir): os.mkdir(triggerDir)
if not os.path.isdir(skymapDir): os.mkdir(skymapDir)
if not os.path.isdir(plotDir): os.mkdir(plotDir)

rm_command = "rm skymap.fits.gz*"
os.system(rm_command)

# Instantiate client
g = GraceDb()
#g = GraceDb(url)
#g = GraceDbBasic()
#g = GraceDbBasic(url)

eventString = get_eventstring(opts)
if opts.doGPSLoop:
    opts.doGPS = True
    while True:
        endtime = Time(datetime.utcnow(), scale='utc')
        starttime = t = Time(endtime.gps - 86400.0, format='gps', scale='utc')
        opts.startGPS = starttime.gps
        opts.endGPS = endtime.gps

        eventString = get_eventstring(opts)
        # REST API returns an iterator
        events = g.events('%s'%eventString)

        download_events(events)
else:
    eventString = get_eventstring(opts)
    print eventString
    # REST API returns an iterator
    events = g.events('%s'%eventString)

    download_events(events) 

